/*
 * Copyright (c) 2004-2007 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2008 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */

#include "ompi_config.h"

#include <string.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <unistd.h>

#if OMPI_BTL_HAVE_SRIO
#include <linux/rio_mport_cdev.h>
#include <linux/rio_regs.h>
#endif


#include "opal/class/opal_bitmap.h"
#include "opal/datatype/opal_convertor.h"
#include "opal/sys/atomic.h"
#include "ompi/mca/btl/btl.h"
#include "ompi/mca/mpool/base/base.h"
#include "btl_srio.h"
#include "btl_srio_frag.h"
#include "pktfwdK2H/mpi_interfaces.h"
#include "ompi/proc/proc.h"
#include <ti/cmem.h>

#define BTL_SRIO_MAX_ROUTING_GROUPS 46
/***************************************************************************************************/
mca_btl_srio_t mca_btl_srio = {
 {
    &mca_btl_srio_component.super,
    0, /* btl_eager_limit */
    0, /* btl_rndv_eager_limit */
    0, /* btl_max_send_size */
    0, /* btl_rdma_pipeline_send_length */
    0, /* btl_rdma_pipeline_frag_size */
    0, /* btl_min_rdma_pipeline_size */
    0, /* btl_exclusivity */
    0, /* btl_latency */
    0, /* btl_bandwidth */
    0, /* btl flags */
    0, /* btl segment size */
    mca_btl_srio_add_procs,
    mca_btl_srio_del_procs,
    NULL,
    mca_btl_srio_finalize,
    mca_btl_srio_alloc,
    mca_btl_srio_free,
    mca_btl_srio_prepare_src,
    mca_btl_srio_prepare_dst,
    mca_btl_srio_send,
    mca_btl_srio_sendi,
    NULL,
    NULL,
    mca_btl_base_dump,
    NULL, /* mpool */
    NULL, /* register error cb */
    mca_btl_srio_ft_event
  }
};
/***************************************************************************************************/
int
mca_btl_srio_module_init (mca_btl_srio_t *srio_module,
                          void *dev)
{
    int rc;
    opal_output_verbose(20, mca_btl_base_output, "binding module %p to device %p", (void *) srio_module, (void *) dev);

    /* copy module defaults (and function pointers) */
    memmove (srio_module, &mca_btl_srio, sizeof (mca_btl_srio_t));

    srio_module->endpoints = NULL;
    mca_btl_srio_component.endpoint_count = 0;
    return OMPI_SUCCESS;
}
/***************************************************************************************************/
int mca_btl_srio_init_ep (int local_rank, int peer_rank, int srio_id, 
                           mca_btl_base_endpoint_t **ep,
                           mca_btl_srio_t *btl,
                           ompi_proc_t *peer_proc) 
{
    mca_btl_base_endpoint_t *endpoint;
    int rc;

    endpoint = calloc(1, sizeof(mca_btl_base_endpoint_t));
    assert (endpoint != NULL);

    endpoint->btl = (void *)btl;
    endpoint->local_rank = local_rank;
    endpoint->peer_rank  = peer_rank; /* peer process to communicate with */
    endpoint->srio_id    = srio_id;
    strncpy(endpoint->hostname, peer_proc->proc_hostname, 80);
    *ep = endpoint;

    return OMPI_SUCCESS;
}
/***************************************************************************************************/
int mca_btl_srio_add_procs( struct mca_btl_base_module_t* btl, 
                            size_t nprocs, 
                            struct ompi_proc_t **procs, 
                            struct mca_btl_base_endpoint_t **peers,
                            opal_bitmap_t* reachability )
{
    int i;
    int srio_iface = 0;
    int dest_cartridge_id, dest_node_id;
    unsigned short srioID_dest;
    size_t ntotal_procs, nlocal_procs;
    mca_btl_srio_t *srio_module = (mca_btl_srio_t *) btl;
    bool first_time_init = (NULL == srio_module->endpoints);
    ompi_proc_t* my_proc; /* pointer to caller's proc structure */
    int rc;
    int list_of_srio_ids[BTL_SRIO_MAX_PROCESS];

    /* get pointer to my proc structure */
    if(NULL == (my_proc = ompi_proc_local())) return OMPI_ERR_OUT_OF_RESOURCE;
    memset (list_of_srio_ids, -1, sizeof(list_of_srio_ids));
    if (NULL == srio_module->endpoints) 
    { /* Allocate memory space for all endpoints */
      (void) ompi_proc_world (&ntotal_procs);

      srio_module->endpoints = calloc (ntotal_procs, sizeof (mca_btl_base_endpoint_t *));
      if (OPAL_UNLIKELY(NULL == srio_module->endpoints)) return OMPI_ERR_OUT_OF_RESOURCE;

      /* initialize own FIFOs */
      /*
       * The receiver initializes all its FIFOs.  All components will
       * be allocated near the receiver.  Nothing will be local to
       * "the sender" since there will be many senders.
       */
      if(nprocs > BTL_SRIO_MAX_PROCESS) {
        opal_output(0, "BTL_SRIO: Maximum number of MPI processes exceeded (%d > %d)\n", nprocs, BTL_SRIO_MAX_PROCESS);
        return OMPI_ERR_OUT_OF_RESOURCE;
      }
    }
   
    for (i = 0; i < nprocs ; ++i) 
    { /* Find my rank */
      struct ompi_proc_t *ompi_proc = procs[i];
      /* check to see if this is me */
      if(my_proc == procs[i]) {
        mca_btl_srio_component.my_rank = ompi_proc->proc_name.vpid;
        break;
      }
    }
    opal_output_verbose(20, mca_btl_base_output, "ADD_PROCS my_rank %d isearch %d\n", mca_btl_srio_component.my_rank, i);
    if(i >= nprocs) return OMPI_ERR_OUT_OF_RESOURCE; /* my_rank not found! */
#if OMPI_BTL_HAVE_SRIO
/*****/
      { /** Create routing table for cartridges in clusters **/
        struct ompi_proc_t *ompi_proc;
        int cartridge_phys_id[BTL_SRIO_MAX_ROUTING_GROUPS] = { 0 };
        unsigned int tot_cartridges = 0, cartridges_bfield[1 + (BTL_SRIO_MAX_ROUTING_GROUPS/32)] = { 0 };
        for(i = 0; i < nprocs; i ++) {
           ompi_proc = procs[i];
           if(SrioHostname2PhysID (mca_btl_srio_component.hostnameList, ompi_proc->proc_hostname, &dest_cartridge_id, &dest_node_id) < 0) continue;
           cartridges_bfield[dest_cartridge_id / 32] |= 1 << (dest_cartridge_id % 32);
        }
        for(i = 0; i < BTL_SRIO_MAX_ROUTING_GROUPS; i ++) {
          if(cartridges_bfield[i / 32] & (1 << (i % 32))) {
            opal_output_verbose(10, mca_btl_base_output, "Cluster cartridge c%d included", i);
            cartridge_phys_id[tot_cartridges ++] = i;
          }
        }
        opal_output_verbose(10, mca_btl_base_output, "Total number of cartridges involved:%d\n", tot_cartridges);
        if(!tot_cartridges) {
          opal_output(0, "No valid cartridges!\n");
          return OMPI_ERR_OUT_OF_RESOURCE;
        }
        opal_output_verbose(10, mca_btl_base_output, "Cluster Routing table creation!");
        if(SrioCreateRoutingTableFromCluster(mca_btl_srio_component.routingTbl, cartridge_phys_id, tot_cartridges) < 0)
        {
          opal_output(0, "Cluster Routing table creation failed!\n");
          return OMPI_ERR_OUT_OF_RESOURCE;
        }
      }
      SrioGetSrioIDFromPhysID(mca_btl_srio_component.routingTbl,
                              mca_btl_srio_component.cartridge_id, mca_btl_srio_component.node_id, &mca_btl_srio_component.srio_id);
      opal_output_verbose(5, mca_btl_base_output, "Local srioID is %d (cart:%d node:%d) tbl=%p\n", mca_btl_srio_component.srio_id,
                          mca_btl_srio_component.cartridge_id, mca_btl_srio_component.node_id, mca_btl_srio_component.routingTbl);
#endif
    nlocal_procs = 0;
    for (i = 0; i < nprocs ; ++i) 
    {
      struct ompi_proc_t *ompi_proc = procs[i];
      uint32_t rem_rank = ompi_proc->proc_name.vpid;
      uint32_t rem_jobid = ompi_proc->proc_name.jobid;

      mca_btl_srio_component.srio_endpoint[i].idx        =  0;
      mca_btl_srio_component.srio_endpoint[i].prev_seqn  = -1;
      mca_btl_srio_component.srio_endpoint[i].srio_id    = -1;
      mca_btl_srio_component.srio_endpoint[i].virt_baseaddr = (char *)mca_btl_srio_component.win_ddr + i * BTL_SRIO_MAX_SEND_SIZE;

      if (OPAL_PROC_ON_LOCAL_NODE(ompi_proc->proc_flags)) {
        /* don't use SRIO to communicate with local procs */
        continue;
      }
 
      if(SrioHostname2PhysID (mca_btl_srio_component.hostnameList, ompi_proc->proc_hostname, &dest_cartridge_id, &dest_node_id) < 0)
      {
        opal_output(0, "Invalid Hostname (%s)", ompi_proc->proc_hostname);
        continue;
      }
      /* Find srio_iface based on cartridge and nodes id-s */
      if(SrioGetSrioIDFromPhysID(mca_btl_srio_component.routingTbl, dest_cartridge_id, dest_node_id, &srioID_dest) < 0)
      {
        opal_output(0, "Invalid destination cartridgeID (%d) or nodeID (%d)", dest_cartridge_id, dest_node_id);
        continue;
        //return OMPI_ERR_OUT_OF_RESOURCE;
      }
      mca_btl_srio_component.srio_endpoint[i].srio_id    = srioID_dest;
      /* Create and Init endpoints */
      rc = mca_btl_srio_init_ep (nlocal_procs /*IN*/, rem_rank /*IN*/, srioID_dest /*IN*/,
                                 peers + i /*OUT*/, (mca_btl_srio_t *) btl, ompi_proc);

      if (OPAL_UNLIKELY(OMPI_SUCCESS != rc)) {
        opal_output(0, "btl/srio error initializing endpoint");
        return rc;
      }
      /* Set the reachable bit */
      rc = opal_bitmap_set_bit (reachability, i);

      /* Store a reference to this peer */
      srio_module->endpoints[rem_rank] = peers[i];
      list_of_srio_ids[i] = srioID_dest;
      nlocal_procs ++;
    }
    mca_btl_srio_component.endpoint_count += nprocs;
    mca_btl_srio_component.nlocal_procs    = nlocal_procs;
    /******************************************************************************************/
    Srio_Type11_setup (mca_btl_srio_component.type11_drv, mca_btl_srio_component.routingTbl,
                       list_of_srio_ids, mca_btl_srio_component.endpoint_count);
    /******************************************************************************************/
    opal_output_verbose (0, mca_btl_base_output, "TIMPIDBG srio_module_add_procs, RANKS: my_rank=%d nprocs=%d nlocal_procs=%d endpoint_count=%d\n", 
       mca_btl_srio_component.my_rank, nprocs, nlocal_procs, mca_btl_srio_component.endpoint_count);
    opal_output_verbose (0, mca_btl_base_output, "TIMPIDBG srio_module_add_procs, list of srio IDs: %d %d %d %d %d %d %d %d\n",
       list_of_srio_ids[0], list_of_srio_ids[1], list_of_srio_ids[2], list_of_srio_ids[3],
       list_of_srio_ids[4], list_of_srio_ids[5], list_of_srio_ids[6], list_of_srio_ids[7]);
    return OMPI_SUCCESS;
}
/***************************************************************************************************/

int mca_btl_srio_del_procs( struct mca_btl_base_module_t* btl, 
                            size_t nprocs,
                            struct ompi_proc_t **procs, 
                            struct mca_btl_base_endpoint_t **peers )
{
    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: srio_module_del_procs!!!\n");
    return OMPI_SUCCESS;
}
/******************************************************************************/
/**
 * MCA->BTL Clean up any resources held by BTL module
 * before the module is unloaded.
 *
 * @param btl (IN)   BTL module.
 *
 * Prior to unloading a BTL module, the MCA framework will call
 * the BTL finalize method of the module. Any resources held by
 * the BTL should be released and if required the memory corresponding
 * to the BTL module freed.
 *
 */
uint32_t btl_srio_probe_point_trip_counter[10] = { 0 };
uint32_t btl_srio_probe_point_start = 0;
uint64_t btl_srio_probe_point_time[10] = { 0 };

int mca_btl_srio_finalize(struct mca_btl_base_module_t* btl)
{
CMEM_AllocParams  alloc_paramsHeap;
int i;
  opal_output_verbose(5, mca_btl_base_output, "srio_module_finalize!!!");
#if 0
  for (i = 0; i < 8; i ++) {
    opal_output_verbose(0, mca_btl_base_output, "PROF(%d) %9u %10.1lf", i, btl_srio_probe_point_trip_counter[i],
       btl_srio_probe_point_trip_counter[i] ? ((double)btl_srio_probe_point_time[i] / (double)btl_srio_probe_point_trip_counter[i]):0);
  }
#endif
  sleep(1);
  /***Release CMEM buffer***/
  alloc_paramsHeap.type = CMEM_HEAP;
#if 0
  if(mca_btl_srio_component.cmem_cached) alloc_paramsHeap.flags = CMEM_CACHED;
  else alloc_paramsHeap.flags = CMEM_NONCACHED;
#else
  alloc_paramsHeap.flags = CMEM_CACHED;
#endif
  alloc_paramsHeap.alignment  = 4;
  CMEM_free(mca_btl_srio_component.virt_cmem_addr, &alloc_paramsHeap);
  /* This will free CMEM buffer allocated in Srio_Type11_init() */
  Srio_Type11_finalize (mca_btl_srio_component.type11_drv);
  CMEM_exit();
  return OMPI_SUCCESS;
}
/******************************************************************************/

/**
 * Allocate a segment.
 *
 * @param btl (IN)      BTL module
 * @param size (IN)     Request segment size.
 */
mca_btl_base_descriptor_t* mca_btl_srio_alloc(
        struct mca_btl_base_module_t* btl,
        struct mca_btl_base_endpoint_t* endpoint,
        uint8_t order,
        size_t size,
        uint32_t flags)
{
    mca_btl_srio_frag_t* frag = NULL;
    int rc;

    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: srio_module_alloc (size=%d)!!!\n", size);
    if(size <= mca_btl_srio.super.btl_eager_limit) {
        MCA_BTL_SRIO_FRAG_ALLOC_EAGER(frag,rc);
    } else if (size <= mca_btl_srio_component.max_frag_size) {
    opal_output_verbose(1, mca_btl_base_output, "TIMPIDBG: srio_module_alloc MAX (size=%d)!!!\n", size);
        MCA_BTL_SRIO_FRAG_ALLOC_MAX(frag,rc);
    }

    if (OPAL_LIKELY(frag != NULL)) {
        frag->segment.seg_len = size;
        frag->base.des_flags = flags;
    }
    return (mca_btl_base_descriptor_t*)frag;
}

/******************************************************************************/
/**
 * Return a segment allocated by this BTL.
 *
 * @param btl (IN)      BTL module
 * @param segment (IN)  Allocated segment.
 */
extern int mca_btl_srio_free(
    struct mca_btl_base_module_t* btl,
    mca_btl_base_descriptor_t* des)
{
    mca_btl_srio_frag_t* frag = (mca_btl_srio_frag_t*)des;
    opal_output_verbose(20, mca_btl_base_output, "BTL_SRIO free!!!\n");
    MCA_BTL_SRIO_FRAG_RETURN(frag);

    return OMPI_SUCCESS;
}
/******************************************************************************/
/**
 * Prepare data for send/put
 *
 * @param btl (IN)      BTL module
 */
struct mca_btl_base_descriptor_t*
mca_btl_srio_prepare_src( struct mca_btl_base_module_t* btl,
                          struct mca_btl_base_endpoint_t* endpoint,
                          mca_mpool_base_registration_t* registration,
                          struct opal_convertor_t* convertor,
                          uint8_t order,
                          size_t reserve,
                          size_t* size,
                          uint32_t flags )
{
    mca_btl_srio_frag_t* frag;
    struct iovec iov;
    uint32_t iov_count = 1;
    size_t max_data = *size;
    int rc;
    unsigned int *data_ptr;
    int jj;    

    opal_convertor_get_current_pointer (convertor, &data_ptr);
    opal_output_verbose(20, mca_btl_base_output, "BTL_SRIO prepare_src EAGER: %d MAX: %d size=%d (btl=%p)", 
          mca_btl_srio_component.eager_limit, mca_btl_srio_component.max_frag_size, reserve + max_data, btl);

    if ( reserve + max_data <= mca_btl_srio_component.eager_limit ) {
        MCA_BTL_SRIO_FRAG_ALLOC_EAGER(frag,rc);
    } else {
        MCA_BTL_SRIO_FRAG_ALLOC_MAX(frag, rc);
    }
    opal_output_verbose(20, mca_btl_base_output, 
        "BTL_SRIO prepare_src CHK_A0 (frag->size=%d) reserve=%d, frag=%p [%d][%d]\n", frag->size, reserve, frag,
         mca_btl_srio_component.frag_cnt[0], mca_btl_srio_component.frag_cnt[1]);

    if( OPAL_UNLIKELY(NULL == frag) ) {
        opal_output(0, "BTL_SRIO  prepare_src FAILED TO ALLOCATE!\n");
        return NULL;
    }
    opal_output_verbose(20, mca_btl_base_output, "BTL_SRIO prepare_src CHK_A1");
    if( OPAL_UNLIKELY(reserve + max_data > frag->size) ) {
        max_data = frag->size - reserve;
    }
    iov.iov_len = max_data;
    iov.iov_base =
        (IOVBASE_TYPE*)(((unsigned char*)(frag->segment.seg_addr.pval)) + reserve);
    opal_output_verbose(20, mca_btl_base_output, "BTL_SRIO prepare_src CHK_B(%p), len=%d\n", iov.iov_base, iov.iov_len);
    rc = opal_convertor_pack(convertor, &iov, &iov_count, &max_data );
    if( OPAL_UNLIKELY(rc < 0) ) {
        opal_output_verbose(20, mca_btl_base_output, "BTL_SRIO prepare_src fail\n");
        MCA_BTL_SRIO_FRAG_RETURN(frag);
        return NULL;
    }
    frag->segment.seg_len = reserve + max_data;

    opal_output_verbose(20, mca_btl_base_output, "BTL_SRIO prepare_src CHK_C");
    frag->base.des_src = &(frag->segment);
    frag->base.des_src_cnt = 1;
    frag->base.order = MCA_BTL_NO_ORDER;
    frag->base.des_dst = NULL;
    frag->base.des_dst_cnt = 0;
    frag->base.des_flags = flags;
    *size = max_data;
    return &frag->base;
}

/**
 * Prepare data for receive.
 */
struct mca_btl_base_descriptor_t*
mca_btl_srio_prepare_dst( struct mca_btl_base_module_t* btl,
                          struct mca_btl_base_endpoint_t* endpoint,
                          mca_mpool_base_registration_t* registration,
                          struct opal_convertor_t* convertor,
                          uint8_t order,
                          size_t reserve,
                          size_t* size,
                          uint32_t flags )
{
    int rc;
    void *ptr;
    mca_btl_srio_frag_t* frag;

    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG:A srio_module_prepare_dst!!!\n");

    MCA_BTL_SRIO_FRAG_ALLOC_MAX(frag, rc);
    if(OPAL_UNLIKELY(NULL == frag)) {
        return NULL;
    }

    frag->segment.seg_len = *size;
    opal_convertor_get_current_pointer( convertor, &ptr );
    frag->segment.seg_addr.lval = (uint64_t)(uintptr_t) ptr;
    
    frag->base.des_src = NULL;
    frag->base.des_src_cnt = 0;
    frag->base.des_dst = (mca_btl_base_segment_t*)&frag->segment;
    frag->base.des_dst_cnt = 1;
    frag->base.des_flags = flags;

    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG:B srio_module_prepare_dst!!!\n");
    return &frag->base;
}
/******************************************************************************/
/**
 * Initiate a send to the peer.
 *
 * @param btl (IN)      BTL module
 * @param peer (IN)     BTL peer addressing
 */
int mca_btl_srio_send( struct mca_btl_base_module_t* btl,
                       struct mca_btl_base_endpoint_t* endpoint,
                       struct mca_btl_base_descriptor_t* des,
                       mca_btl_base_tag_t tag )
{
    mca_btl_srio_t *srio_module = (mca_btl_srio_t *) btl;
    mca_btl_srio_frag_t *frag = (mca_btl_srio_frag_t *)des;
    unsigned int *rmt_values, i;
    unsigned int seg_len_align;
    int rmt_ep;
    void *tmp_fifo;

    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: srio_module_send frag=%p (base=%p size=%lld)!!!\n", frag,
                        frag->segment.seg_addr.pval, frag->segment.seg_len ); 
    /* available header space */
    frag->hdr->len = frag->segment.seg_len;
    /* type of message, pt-2-pt, one-sided, etc */
    frag->hdr->tag = tag;
    frag->endpoint = endpoint;

    seg_len_align = btl_srio_align32((unsigned int)frag->segment.seg_len);
    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: srio_module_send FIFO_WRITE: (%d)->%d, frag_hdr=%p\n", 
                        mca_btl_srio_component.my_rank, endpoint->peer_rank, frag->hdr );

    while(srio_type11_fifo_write (endpoint->srio_id, 
                                  frag->hdr, sizeof(mca_btl_srio_hdr_t),
                                  NULL, 0, frag->segment.seg_addr.pval, seg_len_align) != OMPI_SUCCESS)
    {
       opal_output_verbose(20, mca_btl_base_output,mca_btl_base_output, "BTL_SRIO no room in FIFO (send)!\n");
    }
    mca_btl_srio_frag_complete (frag);
    return 1;
}

/**
 * Initiate an inline send to the peer. If failure then return a descriptor.
 *
 * @param btl (IN)      BTL module
 * @param peer (IN)     BTL peer addressing
 */
int mca_btl_srio_sendi( struct mca_btl_base_module_t* btl,
                      struct mca_btl_base_endpoint_t* endpoint,
                      struct opal_convertor_t* convertor,
                      void* header,
                      size_t header_size,
                      size_t payload_size,
                      uint8_t order,
                      uint32_t flags,
                      mca_btl_base_tag_t tag,
                      mca_btl_base_descriptor_t** descriptor )
{
   size_t length = (header_size + payload_size);
   mca_btl_srio_frag_t* frag;
   int rc, rmt_ep;
   unsigned int seg_len_align;
   void *data_ptr = NULL;
   char *ptr;
   char host_name[80];
   void *tmp_fifo;
   int *tmp_rmt_msg_idx;

    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: srio_module_sendi A (header_size=%d payload_size=%d tag=%d)\n", 
                        header_size, payload_size, tag);
    /* this check should be unnecessary... turn into an assertion? */
    assert (length < mca_btl_srio_component.eager_limit);
    assert (0 == (flags & MCA_BTL_DES_SEND_ALWAYS_CALLBACK));
    //frag->base.des_flags = flags | MCA_BTL_DES_FLAGS_BTL_OWNERSHIP;   /* why do any flags matter here other than OWNERSHIP? */
 
    /* allocate a fragment, giving up if we can't get one */
    /* note that frag==NULL is equivalent to rc returning an error code */
    MCA_BTL_SRIO_FRAG_ALLOC_EAGER(frag, rc);
    if( OPAL_UNLIKELY(NULL == frag) ) {
        *descriptor = NULL;
        return rc;
    }
    /* fill in fragment fields */
    frag->segment.seg_len = length;
    frag->hdr->len = length;
    frag->hdr->tag = tag;
    /* bypassing the convertor may speed things up a little */
    if(payload_size) {
      opal_convertor_get_current_pointer (convertor, &data_ptr);
      seg_len_align = btl_srio_align32 (payload_size);
    } else seg_len_align = 0;
    while(srio_type11_fifo_write (endpoint->srio_id,
                                  frag->hdr, sizeof(mca_btl_srio_hdr_t), 
                                  header, header_size, 
                                  data_ptr, seg_len_align ))
    {
      opal_output_verbose(20, mca_btl_base_output, "BTL_SRIO no room in FIFO (send)!\n");
    }
    mca_btl_srio_frag_complete (frag);
    return OMPI_SUCCESS;
}
/******************************************************************************/
int mca_btl_srio_ft_event(int state) {
    if(OPAL_CRS_CHECKPOINT == state) {
        ;
    }
    else if(OPAL_CRS_CONTINUE == state) {
        ;
    }
    else if(OPAL_CRS_RESTART == state) {
        ;
    }
    else if(OPAL_CRS_TERM == state ) {
        ;
    }
    else {
        ;
    }
    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: srio_module_ft_event???!!!\n");
    return OMPI_SUCCESS;
}
/******************************************************************************/
