/**
 *   @file  multicoreLoopback.c
 *
 *   @brief   
 *      This is an example application which showcases how the SRIO Driver
 *      API can be used in a multiple core environment. 
 *
 *      The test suite sends/receives data as follows:-
 *
 *      CORE 1 --> CORE 2 --> CORE 3 --> CORE 0 --> CORE 1
 *
 *      At each step the received data is validated (Payload and Received
 *      Identifiers).
 *
 *      The goal of the test suite is to showcase that the SRIO Driver API
 *      are multi-core safe.
 *
 *  \par
 *  ============================================================================
 *  @n   (C) Copyright 2009-2014, Texas Instruments, Inc.
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  \par
*/
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sched.h>
#include<signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stddef.h>
#include <time.h>

/* CPPI/QMSS Include Files. */
#include <ti/drv/cppi/cppi_drv.h>
#include <ti/drv/qmss/qmss_drv.h>
#include <ti/drv/qmss/qmss_qm.h>
#include <ti/drv/qmss/qmss_srioRtr.h>
/* CSL Include Files */
#include <ti/csl/csl_chip.h>
#include <ti/csl/csl_psc.h>
#include <ti/csl/csl_pscAux.h>
#include <ti/csl/csl_srioAux.h>

#include <ti/csl/csl_srioAuxPhyLayer.h>

#include <ti/csl/csl_serdes_srio.h>

/* Device specific include */
#include "fw_test.h"
#include "fw_mem_allocator.h"
#include "linuxutil.h"
#include "srio_test.h"
#include "rtr_sriodrv.h"
#include "srioPlatCfg.h"
#include "srio_osal.h"
#include "../pktfwdK2H/mpi_interfaces.h"


volatile int ctrl_c_received=0;

#define TEST_PORTS                  4
/**********************************************************************
 ************************** LOCAL Definitions *************************
 **********************************************************************/


/* Number of cores for which the test is being executed. */
#define NUM_CORES                   srio_EXAMPLE_NUM_CORES

#if NUM_CORES > 1
/* Defines the core number responsible for system initialization. */
#define CORE_SYS_INIT               1
#else
#define CORE_SYS_INIT               0
#endif


/* SRIO Application Hardware Semaphore. */
#define SRIO_APP_HW_SEM             5

/**********************************************************************
 ************************** Global Variables **************************
 **********************************************************************/

/* Memory used for the descriptors. */
Uint8   *host_region_terminate;
Uint8   *host_region_credit;
Uint8   *host_region_forward;


Qmss_InitCfg   qmssInitConfig;

Qmss_Result    qmssRegion_terminate;
Qmss_Result    qmssRegion_credit;
Qmss_Result    qmssRegion_forward;


/* Global Varialble which keeps track of the core number executing the
 * application. */
uint32_t        coreNum = 0xFFFF;


/* These are the device identifiers used used in the TEST Application */
const uint32_t DEVICE_ID1_16BIT    = 0xBEEF;
const uint32_t DEVICE_ID1_8BIT     = 0xAB;
const uint32_t DEVICE_ID2_16BIT    = 0x4560;
const uint32_t DEVICE_ID2_8BIT     = 0xCD;
const uint32_t DEVICE_ID3_16BIT    = 0x1234;
const uint32_t DEVICE_ID3_8BIT     = 0x12;
const uint32_t DEVICE_ID4_16BIT    = 0x5678;
const uint32_t DEVICE_ID4_8BIT     = 0x56;

Qmss_QueueHnd           qpendRxQueHnd;
int                     qpendFd;
int                     srioFd;
Srio_DrvHandle          isr_hSrioDriver;

/**********************************************************************
 ************************* Extern Definitions *************************
 **********************************************************************/

extern uint32_t malloc_counter;
extern uint32_t free_counter;

extern int32_t SrioDevice_init (CSL_SrioHandle hSrio, uint32_t srioSerdesVAddr);
extern int32_t SrioDevice_init_lite (CSL_SrioHandle hSrio, uint32_t srioSerdesVAddr);
extern int32_t SrioDevice_deinit (CSL_SrioHandle hSrio);
extern int32_t SrioDevice_deinit_lite (CSL_SrioHandle hSrio);

extern void* Osal_srioDataBufferMalloc(uint32_t numBytes);

/* QMSS device specific configuration */
extern Qmss_GlobalConfigParams  qmssGblCfgParams;

/* OSAL Data Buffer Memory Initialization. */
extern int32_t Osal_dataBufferInitMemory(uint32_t dataBufferSize);


void sig_handler(int signo)
{
  if (signo == SIGINT) {
    //printf("received SIGINT\n");
    ctrl_c_received=1;
    printf("\n Termination signal Ctrl-C received....\n");fflush(stdout);
  }
}

/*** Time measurement functions ***/
uint64_t  get_time_us() {
   struct timespec time_now;
   clock_gettime(CLOCK_MONOTONIC, &time_now);
   return ((uint64_t)time_now.tv_sec*1000000  +  (uint64_t)time_now.tv_nsec / 1000);
}


static inline int64_t clock_diff (struct timespec *start, struct timespec *end)
{
    return ( ((uint64_t)end->tv_sec -(uint64_t)start->tv_sec) * 1000000000
             + (uint64_t)end->tv_nsec - (uint64_t )start->tv_nsec );
}

static inline uint64_t  clock_ns (struct timespec *time_now) {
   return ((uint64_t)time_now->tv_sec*1000000000  +  (uint64_t)time_now->tv_nsec);
}
/************************************************************/
typedef struct {
  int tx_terminate_queues;
  int txFree_terminate_queues;
  int tx_srio_credit_queues;
  int txFree_credit_queues;
  int rx_credit_queues;
  int rxFree_credit_queues;
  int tx_srio_queues;
  int rx_completed_forward_queues;
  int rx_forward_return_queues;
  int rxFree_forward_queues;
} djdbg_report_queue_count_per_port_t;
typedef struct {
  int rx_terminate_queue;
  int rxFree_terminate_queue;
//  int errorQueue;
  int garbage_len_queue;
  int garbage_tout_queue;
  int garbage_retry_queue;
  int garbage_transerr_queue;
  int garbage_prog_queue;
  int garbage_ssize_queue;

  djdbg_report_queue_count_per_port_t ports[4];
} djdbg_report_queue_count_t;

volatile djdbg_report_queue_count_t djdbg_report_q;
volatile int djdbgq[64];
/************ Terminiation Queues ******************/
void djdbg_update_queues(void)
{
int i, j;
j = 0;
djdbgq[j ++] = Qmss_getQueueEntryCount(Sriotype11APIsContext.srioQueues.rx_terminate_queue);
djdbgq[j ++] = Qmss_getQueueEntryCount(Sriotype11APIsContext.srioQueues.rxFree_terminate_queue);
//djdbgq[j ++] = Qmss_getQueueEntryCount(Sriotype11APIsContext.srioQueues.errorQueue);
for (i = 905; i <= 910; i ++)
{
  djdbgq[j ++] = Qmss_getQueueEntryCount((Qmss_QueueHnd)i); 
}

for (i = 0; i < 4; i ++) {
  djdbgq[j ++] = Qmss_getQueueEntryCount(Sriotype11APIsContext.srioQueues.tx_terminate_queues[i]); 
  djdbgq[j ++] = Qmss_getQueueEntryCount(Sriotype11APIsContext.srioQueues.txFree_terminate_queues[i]);
  djdbgq[j ++] = Qmss_getQueueEntryCount(Sriotype11APIsContext.srioQueues.tx_srio_credit_queues[i]); 
  djdbgq[j ++] = Qmss_getQueueEntryCount(Sriotype11APIsContext.srioQueues.txFree_credit_queues[i]);
  djdbgq[j ++] = Qmss_getQueueEntryCount(Sriotype11APIsContext.srioQueues.rx_credit_queues[i]); 
  djdbgq[j ++] = Qmss_getQueueEntryCount(Sriotype11APIsContext.srioQueues.rxFree_credit_queues[i]);
  djdbgq[j ++] = Qmss_getQueueEntryCount(Sriotype11APIsContext.srioQueues.tx_srio_queues[i]);
  djdbgq[j ++] = Qmss_getQueueEntryCount(Sriotype11APIsContext.srioQueues.rx_completed_forward_queues[i]);
  djdbgq[j ++] = Qmss_getQueueEntryCount(Sriotype11APIsContext.srioQueues.rx_forward_return_queues[i]); 
  djdbgq[j ++] = Qmss_getQueueEntryCount(Sriotype11APIsContext.srioQueues.rxFree_forward_queues[i]);

}
memcpy (&djdbg_report_q, &djdbgq, sizeof(djdbg_report_queue_count_t));
}

void djdbg_report_queues(int tag_id)
{
int i, j;
  /************ Terminiation Queues ******************/
j = 0;
Srio_osalLogC ("QDEPTH:%d   rx_terminate_queue     %4d", tag_id, djdbgq[j ++] );
Srio_osalLogC ("QDEPTH:%d   rxFree_terminate_queue %4d", tag_id, djdbgq[j ++] );
//Srio_osalLogC ("QDEPTH:%d   error_queue            %4d", tag_id, djdbgq[j ++] );
Srio_osalLogC ("QDEPTH:%d   garbage_len_queue      %4d", tag_id, djdbgq[j ++]);
Srio_osalLogC ("QDEPTH:%d   garbage_tout_queue     %4d", tag_id, djdbgq[j ++]);
Srio_osalLogC ("QDEPTH:%d   garbage_retry_queue    %4d", tag_id, djdbgq[j ++]);
Srio_osalLogC ("QDEPTH:%d   garbage_transerr_queue %4d", tag_id, djdbgq[j ++]);
Srio_osalLogC ("QDEPTH:%d   garbage_prog_queue     %4d", tag_id, djdbgq[j ++]);
Srio_osalLogC ("QDEPTH:%d   garbage_ssize_queue    %4d", tag_id, djdbgq[j ++]);

for (i = 0; i < 4; i ++) {
  Srio_osalLogC ("QDEPTH:%d %d tx_terminate_queue     %4d", tag_id, i, djdbgq[j ++] );
  Srio_osalLogC ("QDEPTH:%d %d txFree_terminate_queue %4d", tag_id, i, djdbgq[j ++] );
  Srio_osalLogC ("QDEPTH:%d %d tx_srio_credit_queue   %4d", tag_id, i, djdbgq[j ++] );
  Srio_osalLogC ("QDEPTH:%d %d txFree_credit_queue    %4d", tag_id, i, djdbgq[j ++] );
  Srio_osalLogC ("QDEPTH:%d %d rx_credit_queue        %4d", tag_id, i, djdbgq[j ++] );
  Srio_osalLogC ("QDEPTH:%d %d rxFree_credit_queues   %4d", tag_id, i, djdbgq[j ++] );
  Srio_osalLogC ("QDEPTH:%d %d tx_srio_queue          %4d", tag_id, i, djdbgq[j ++] );
  Srio_osalLogC ("QDEPTH:%d %d rx_compltforward_queue %4d", tag_id, i, djdbgq[j ++] );
  Srio_osalLogC ("QDEPTH:%d %d rx_forwardreturn_queue %4d", tag_id, i, djdbgq[j ++] );
  Srio_osalLogC ("QDEPTH:%d %d rxFree_forward_queue   %4d", tag_id, i, djdbgq[j ++] );
}

}


int printDataTransferStats(remoteInfo_t *remoteInfo)
{
    int i;
    uint64_t timediff_ns;
	while(!ctrl_c_received) {
	for(i=0;i<remoteInfo->num_entries;i++) {
	  if(remoteInfo->entries[i].print_send_stats) {

   	   printf("\n[SEND]    %d packets sent to # %02d.    Total of %d packets sent so far to # %02d\n",
    	                   remoteInfo->entries[i].print_num_packets_sent,
    	                   remoteInfo->entries[i].remote_id,
            			   remoteInfo->entries[i].print_total_num_packets_sent,
            			   remoteInfo->entries[i].remote_id);fflush(stdout);
   	   remoteInfo->entries[i].print_send_stats=0;

      }

	  if(remoteInfo->entries[i].print_receive_stats) {
		  timediff_ns = clock_diff (&remoteInfo->entries[i].recv_start, &remoteInfo->entries[i].recv_finish);
                 printf("\n[RECEIVE] %d packets received from # %02d. Total of %d packets received so far from # %02d,after %lld ms",
           	    		 remoteInfo->entries[i].print_num_packets_received,
           	    		remoteInfo->entries[i].remote_id,
           	    		remoteInfo->entries[i].print_total_num_packets_received,
           	    		remoteInfo->entries[i].remote_id,timediff_ns/1000000);fflush(stdout);

                 remoteInfo->entries[i].print_receive_stats=0;
	     }
      }
	}
	printf("\n Exiting the thread to print stats\n");
	return 0;

}


#define TEST_TIMEOUT 100000000
#define TEST_NUM_PACKETS 900
#define TEST_NUM_BURST_PACKETS 100
#define TEST_SYNC_THRESHOLD 1000000 /* Stop for sync after sending 1 million */
#define TEST_PRINT_THRESHOLD 100000 /* Print out once in every 100,000 packets */

static Int32 send_receive_Test (Srio_DrvHandle hSrioDrv, Role_e Role,int my_id, remoteInfo_t *remoteInfo)
{
    Srio_SockHandle         srioSocket;
    Srio_SockHandle         srioSocket_perPort[4]={NULL,NULL,NULL,NULL};/* Tx Sockets Per port */
    int num_sockets;
    Srio_SockBindAddrInfo   bindInfo;
    uint8_t*                txData;
    Srio_SockAddrInfo       to;
    Srio_SockAddrInfo       from;
    int                     idx;
    Int32                   num_bytes;
    uint8_t*                ptr_rxDataPayload;
    Int32                   sendToCore;
    Int32                   recvFromCore;
    Srio_DrvBuffer          hDrvBuffer;
    uint32_t                allocatedLen;    
    uint16_t                coreDeviceID[NUM_CORES];
    char                    waitName[32];
    CSL_SrioHandle hSrio = (CSL_SrioHandle)fw_srioCfgVaddr;
    int val,i;
    Srio_txQueueCfg        txQ;
    uint8_t *payload;
    uint32_t qOffset;
    int priority=0,socketno,remote_id,remote_no,port;
    Cppi_Desc *ptrDesc;
    unsigned long count;
    int packet_no,iter;
    uint8_t rcvd_pkt_id;
    int remote_idx=0,remoteinfo_index=0;
    uint8_t id_to_index[256];/* Maps the remote id to index within remoteInfo[] for easier lookup */
    pthread_t print_stats_thread;
    uint64_t time_start, time_now;

    memset(id_to_index,0xff,256*sizeof(uint8_t));

    /* Map remote_id to the index of remoteInfo->entries[] for easier lookup on receiveing packets
     * */
    for(i=0;i<remoteInfo->num_entries;i++) {
      id_to_index[remoteInfo->entries[i].remote_id]= i;
      printf("Mapped id_to_index[%d] to %d\n",remoteInfo->entries[i].remote_id,i);fflush(stdout);
    }

    if (signal(SIGINT, sig_handler) == SIG_ERR) {
      printf("\ncan't catch SIGINT\n");
      return -1;
    }

     /* We can proceed with the data transmission & reception tests only after all the
     * cores have created and bound their SRIO sockets. Using RM name server as barrier.
     */
#ifdef MANUAL_RUN
    printf("\n*************  Ready ************");
    printf("\n Press any key to continue to start transfer..... ");
    printf("\n********************************");

    getchar();
#else
    sleep(1);
#endif

    /* Is this the starting core? */ 
    if (Role == ROLE_SENDER || Role == ROLE_RECEIVER)
    {
    	 /*	while (1)
    	 	   if ( pop(credits debug) )
    	 	      push srio

    	 	   if ( pop(TX-free)
    	 	      PUSH firmware host
    	 	      sends++
    	 	      if (sends % 1000) printf ("sent %d\n", sends);
    	 	   if ( pop (RX))
    	 	      PUSH return
    	 	      receives++
    	 	      if (receives% 1000) printf ("receives %d\n", receives);
    	 */


    	   for(i=0;i<remoteInfo->num_entries;i++) {
           	    srioSocket=remoteInfo->entries[i].socket;
                remote_id=remoteInfo->entries[i].remote_id;
    		   /* Send a synchronization packet to everyone */
    	   		hDrvBuffer = Srio_Type11_AcquireTX (hSrioDrv, (void **)&txData, &allocatedLen, remote_id);
     			  /*  -------- MPI RESPONSIBILITY START -------- */
     	   		    txData[0]=remote_id;/* Destination ID */
     	   		    txData[1]=my_id;/* Source ID */
     	   		    txData[2]=0; /* 2K Message Sequence number */
     	   		    txData[3]=0;/* Total number of 2K messages */
    			  /*  -------- MPI RESPONSIBILITY END  -------- */

   	    		if (hDrvBuffer != NULL) {
   	               if(Srio_Type11_SendTX (hDrvBuffer, hDrvBuffer,remote_id,allocatedLen)<0) {
   	                  printf("\n Srio_Type11_SendTX failed!!");
   	                  goto cleanup_exit; // ?? Or should we just continue?
                    }
   	               printf("\n Sent synch packet to # %d \n",remote_id);fflush(stdout);

   	    		}
    	   }




    	   for(i=0;i<remoteInfo->num_entries;i++) {
    	    	clock_gettime(CLOCK_MONOTONIC, &(remoteInfo->entries[i].recv_start));

    	    	remoteInfo->entries[i].num_packets_sent=0;
    	    	remoteInfo->entries[i].num_packets_received=0;
    	    	remoteInfo->entries[i].total_num_packets_sent=0;
    	    	remoteInfo->entries[i].total_num_packets_received=0;
    	    	remoteInfo->entries[i].sync_packet_received=0;
    	    	remoteInfo->entries[i].print_receive_stats=0;
    	    	remoteInfo->entries[i].print_send_stats=0;
    	    }
    	   if(pthread_create(&print_stats_thread, NULL, printDataTransferStats, remoteInfo)) {

    	     fprintf(stderr, "Error creating thread\n");
    	     goto cleanup_exit;

    	   }

        time_start = get_time_us();
    	remote_idx=0;
    	/* Poll through the list of remote senders/receivers, and send & receive packet from them */
    	while(!ctrl_c_received) { /* while(1) */
    		if(remoteInfo->num_entries==0) {
    			continue;
    		}
            remote_id=remoteInfo->entries[remote_idx].remote_id;

            if(remoteInfo->entries[remote_idx].sync_packet_received)
            {/* Send data only after the first synch packet is received from this sender */

            	hDrvBuffer = Srio_Type11_AcquireTX (hSrioDrv, (void **)&txData, &allocatedLen, remote_id); /* returns cppi_desc */

    		  if (hDrvBuffer != NULL) {
    			   /* MPI Does this. Since No MPI, we need to do this */

    			  /*  -------- MPI RESPONSIBILITY START -------- */
   	   		    txData[0]=remote_id;/* Destination ID */
   	   		    txData[1]=my_id;/* Source ID */
   	   		    txData[2]=0; /* 2K Message Sequence number */
   	   		    txData[3]=0;/* Total number of 2K messages */
  			  /*  -------- MPI RESPONSIBILITY END  -------- */

               if(Srio_Type11_SendTX (hSrioDrv, hDrvBuffer,remote_id,allocatedLen)<0) {
                  printf("\n Srio_Type11_SendTX failed!!");
                  goto cleanup_exit; // ?? Or should we just continue?
               }
               remoteinfo_index=id_to_index[remote_id]; /* Index within the remoteInfo_t structure */
               remoteInfo->entries[remoteinfo_index].num_packets_sent++;
               remoteInfo->entries[remoteinfo_index].total_num_packets_sent++;

               if((remoteInfo->entries[remoteinfo_index].total_num_packets_sent % TEST_SYNC_THRESHOLD)==0) {
            	   remoteInfo->entries[remoteinfo_index].sync_packet_received=0;/* Turning the sync off, to stop sending and thereby ensuring
            	      the remote end is not choked receiving packets from me, and giving a chance to send packets */
               }

               if(remoteInfo->entries[remoteinfo_index].num_packets_sent == TEST_PRINT_THRESHOLD) {
            	   remoteInfo->entries[remoteinfo_index].print_send_stats=1;
            	   remoteInfo->entries[remoteinfo_index].print_num_packets_sent=remoteInfo->entries[remoteinfo_index].num_packets_sent;
            	   remoteInfo->entries[remoteinfo_index].print_total_num_packets_sent=remoteInfo->entries[remoteinfo_index].total_num_packets_sent;
            	   remoteInfo->entries[remoteinfo_index].num_packets_sent =0;
               }

    		  }
            }
    		/* Now receive if any */
    		ptrDesc=(Cppi_Desc *)Srio_Type11_AcquireRX(hSrioDrv, (void **)&payload, &num_bytes);
            if(ptrDesc) {
              /* Find where this packet is from by looking at the header of the payload */
              rcvd_pkt_id=payload[1];

              Srio_MessageRelease(hSrioDrv, ptrDesc); /* move descriptor and buffer to free pool */

              remoteinfo_index=id_to_index[rcvd_pkt_id]; /* Index within the remoteInfo_t structure */
              if(id_to_index[rcvd_pkt_id]==0xff) {
            	 printf("Unexpected packet received from %d\n",rcvd_pkt_id);
            	 continue;
              }
              if(remoteInfo->entries[remoteinfo_index].sync_packet_received==0) {
            	  remoteInfo->entries[remoteinfo_index].sync_packet_received=1;

              }
              remoteInfo->entries[remoteinfo_index].num_packets_received++;
              remoteInfo->entries[remoteinfo_index].total_num_packets_received++;

              if(remoteInfo->entries[remoteinfo_index].num_packets_received == TEST_PRINT_THRESHOLD) {
            	  clock_gettime(CLOCK_MONOTONIC, &(remoteInfo->entries[remoteinfo_index].recv_finish));
            	   remoteInfo->entries[remoteinfo_index].print_receive_stats=1;
            	   remoteInfo->entries[remoteinfo_index].print_num_packets_received=remoteInfo->entries[remoteinfo_index].num_packets_received;
            	   remoteInfo->entries[remoteinfo_index].print_total_num_packets_received=remoteInfo->entries[remoteinfo_index].total_num_packets_received;

           	  remoteInfo->entries[remoteinfo_index].num_packets_received=0;
           	  clock_gettime(CLOCK_MONOTONIC, &(remoteInfo->entries[remoteinfo_index].recv_start));
              }

    		}

        	remote_idx++;
            if(remote_idx==remoteInfo->num_entries) {
              remote_idx=0;
            }
            time_now = get_time_us();
            if((time_now - time_start) > 30000000) ctrl_c_received = 1;
     	 }

    	/* wait for the second thread to finish */
    	if(pthread_join(print_stats_thread, NULL)) {

    	   fprintf(stderr, "Error joining thread\n");
    	   goto cleanup_exit;
    	}

       printf("\n !!! SEND/RECEIVE COMPLETE !!!!!!!!.\n");fflush(stdout);

    } else if(Role == ROLE_ROUTER){
      printf("\n Configured as ROUTER: Do Nothing.......\n");
    }

cleanup_exit:

#ifdef MANUAL_RUN
  printf("\n*************  Exit ************");
  printf("\n Press any key to continue Exiting ..... ");
  printf("\n********************************");
  getchar();
#endif


  printf("\n ----------- Data Transfer Summary -------------\n");
  for(i=0;i<remoteInfo->num_entries;i++) {

   	printf("To   DestID= %02d, Total Packets sent    = %d\n",
   			remoteInfo->entries[i].remote_id,
   			remoteInfo->entries[i].total_num_packets_sent);

   	printf("From DestID= %02d, Total Packets Received = %d\n",
   			remoteInfo->entries[i].remote_id,
   			remoteInfo->entries[i].total_num_packets_received);
   	printf("\n");
  }

    printSrioRtrStats();

    /* We are done with the test. */
    if (errorCount == 0)
    {
        System_printf ("Debug: MultiNode Test Passed\n");
    }
    return 0;
}

/**
 *  @b Description
 *  @n  
 *      This is the Multicore Test task, invoked from main()
 *
 *  @retval
 *      Not Applicable.
 */
extern Srio_DrvHandle Srio_start_machu (Srio_DrvConfig* ptr_cfg);



//#define CARTRIDGE_HIJACK_VALUE 1
/*******************************************************************************************/
/*******************************************************************************************/
int find_my_cartridge_and_node_id(int *cartridge_id, int *node_id) {

  FILE * node_id_detect, *cart_id_detect;
  char node_id_buff[4],cart_id_buff[4];
  unsigned short srio_id;

  node_id_detect = fopen("/proc/device-tree/sl1500/node/id", "rb");
  if(node_id_detect)
  {
     memset(node_id_buff, 0, 4);
     fread(node_id_buff, 1, 3, node_id_detect);
     *node_id = strtoul(node_id_buff, (char **)NULL, 10);
     fclose (node_id_detect);

  }   else {

	  printf("\n The file /proc/device-tree/sl1500/node/id not found !!! ");
      return -1;
  }



  cart_id_detect = fopen("/proc/device-tree/sl1500/cartridge/id", "rb");
  if(cart_id_detect)
  {
     memset(cart_id_buff, 0, 4);
     fread(cart_id_buff, 1, 3, cart_id_detect);
     *cartridge_id = strtoul(cart_id_buff, (char **)NULL, 10);
     fclose (cart_id_detect);

  }else {

	  printf("\n The file /proc/device-tree/sl1500/cartridge/id not found !!! ");
      return -1;
  }
#ifdef  CARTRIDGE_HIJACK_VALUE
          if(*cartridge_id==45) {
        	  *cartridge_id=CARTRIDGE_HIJACK_VALUE;
          }
          printf("\n Cartridge ID hijacked to %d!!",CARTRIDGE_HIJACK_VALUE);
#endif

  return 0;
}

/**
 *  @b Description
 *  @n  
 *     Collect all the command line arguments or provide how-to info
 *     
 *  @retval
 *      Not Applicable.
 */

// #define AB_TEST 1
//#define BC_TEST 1
//#define ABC_TEST 1

/**********************************************************************************************************/
int  form_routingTable(void *base, int cartridge_phys_ids[],int num_cartridge_phys_ids)
{
    int i,ret;

    printf("\n Enabling Cartridges:  ");
    for(i=0;i<num_cartridge_phys_ids;i++) {
      printf("\n c%d",cartridge_phys_ids[i]);
    }
    printf("\n");

    if(ret=SrioCreateRoutingTableFromCluster(base,cartridge_phys_ids,num_cartridge_phys_ids)<0) {
	   printf("\n Routing Table Creation Failed");
    }
   return ret;
}
/**********************************************************************************************************/
void * create_routingStructures(char *topology_file_str)
{
    void *base;
    int i;

    printf("\nCreating Routing structures, based on Topology"); fflush(stdout);
    base = SrioRoutingStructuresCreate(topology_file_str);
    if (base==NULL) {
      printf("\nCreating Failed, returning...");
    }
 return base;
}


/**********************************************************************************************************/
/**********************************************************************************************************/
int multinode_test_main(void *args)
{
    Srio_InitConfig srioInitCfg;
    char            waitName[32];
    int             i,num_remote_entries=0;
    char **argv= (char **)args;
    int port,priority;
    unsigned short remote_id,my_id;
    char role_string[3][15]={"Sender", "Receiver","Router"};
    char topology_file_str[]="/etc/cluster/srio_topology.bin";
    int num_cartridge_ids=0;
    int cartridge_ids[45],cartridge_id;
    int argc=1,remote_cartridge_id,remote_node_id;
    int my_cartridge_id,my_node_id;
    Role_e nodeRole;
    void *SRIORunTimeStructures;
    Srio_DrvHandle srioDrvHandle;
    void *routingTblStructures;
    int list_of_srio_ids[256],num_of_srio_ids=0;
    remoteInfo_t remoteInfo;


    System_printf ("***********************************\n");
    System_printf ("******  Multi Node Testing (Init Arguments now..)  *******\n");
    System_printf ("***********************************\n");

    /* --- MPI RESPONSIBILITY BEGIN ---*/
    printf ("\nDJDJBG... MPI InitTime: Create Routing Structures \n"); fflush(stdout);
    /* Initialize SRIO Routing structures with the cluster information */
    SRIORunTimeStructures = create_routingStructures(topology_file_str);
    /* --- MPI RESPONSIBILITY END ---*/

    /***************************************** INPUT PARSING *************************************************/
    for(i=0;i<15;i++) {
    	printf("\n Arg[%d] = %s",i,argv[i]);
      if(argv[i]==NULL) {
        argc=i;
        break;
      }
    }
    printf("\n Argc = %d\n",argc);

    num_cartridge_ids=0;
    /* Find the cartridges for this cluster */
    for(i=1;i<argc;i++)
    {
    	if((strcmp(argv[i],"sender")==0)   ||
    	   (strcmp(argv[i],"receiver")==0) ||
           (strcmp(argv[i],"router")==0)) {
           break;
        } else {
           sscanf(argv[i],"c%d",&cartridge_id);
#ifdef CARTRIDGE_HIJACK_VALUE
          if(cartridge_id==45) {
        	  cartridge_id=CARTRIDGE_HIJACK_VALUE;
          }
          printf("\n Cartridge ID hijacked to %d !!",CARTRIDGE_HIJACK_VALUE);
#endif
           cartridge_ids[num_cartridge_ids]=cartridge_id;
           printf("INPUT: Cartridge C%d selected\n",cartridge_ids[num_cartridge_ids]);
           num_cartridge_ids++;
       }
    }
printf ("\nDJDJBG... First pass scan done (%d)\n", i); fflush(stdout);

    /* Find the role for this cluster */
    if(strcmp(argv[i],"sender")==0) {
      /* Node A (#3) */
      nodeRole=ROLE_SENDER;
    } else if (strcmp(argv[i],"receiver")==0) {
      nodeRole=ROLE_RECEIVER;
    } else if (strcmp(argv[i],"router")==0) {
      /* Node B (#2) */
      nodeRole=ROLE_ROUTER;
    } else {
      System_printf ("INPUT: Undefined Role!! (should be sender/receiver/router) Quitting..\n");
      return;
    }
    i++;
    num_remote_entries=0;
    Role_global = nodeRole;
    System_printf ("INPUT:  Role = %s\n", role_string[nodeRole]);
    priority=0; /* Hardcoded for now */


    num_remote_entries=0;
    for(;i<argc;i++)
    {
        if(argv[i]==NULL) {
     	  break;
        } else {
          sscanf(argv[i],"c%dn%d",&remote_cartridge_id,&remote_node_id);
          printf("Remote: String: %s,  remote_cartridge_id=%d, remote_node_id=%d",argv[i],remote_cartridge_id,remote_node_id); fflush(stdout);
#ifdef CARTRIDGE_HIJACK_VALUE
          if(remote_cartridge_id==45) {
        	  remote_cartridge_id=CARTRIDGE_HIJACK_VALUE;
          }
          printf("\n Remote Cartridge ID hijacked to %d !!",CARTRIDGE_HIJACK_VALUE);

#endif
          if(SrioGetSrioIDFromPhysID(SRIORunTimeStructures,
        		  remote_cartridge_id, remote_node_id, &remote_id)<0){
        	  printf("\n SrioGetSrioIDFromPhysID Failed!!\n");
        	  return -1;
          }
          printf("\nSrioGetSrioIDFromPhysID %s = %d\n",argv[i],remote_id); fflush(stdout);

          remoteInfo.entries[num_remote_entries].remote_id=remote_id;

          System_printf ("INPUT: [%d] To remote Id %d\n",
        		                        num_remote_entries,
         		                       remoteInfo.entries[num_remote_entries].remote_id);
          num_remote_entries++;
        }
     }
     remoteInfo.num_entries=num_remote_entries;

    /***************************** END OF INPUT PARSING **************************************************/



        /********************************* INIT TIME-1 (Done by MPI)********************************************/
        printf ("\nDJDJBG... About to find_my_node\n"); fflush(stdout);
        find_my_cartridge_and_node_id(&my_cartridge_id,&my_node_id);
        printf ("\nDJDJBG... My cartridge id=%d, My Node id=%d\n",my_cartridge_id,my_node_id); fflush(stdout);

        if(SrioGetSrioIDFromPhysID(SRIORunTimeStructures,my_cartridge_id,my_node_id,&my_id)<0) {
           printf("\n ERROR!! Failed to get my own node id ");
      	   return -1;
        }
        printf ("\nDJDJBG... My cartridge id=%d, My Node id=%d, My SRIO ID = %d\n",my_cartridge_id,my_node_id,my_id); fflush(stdout);
         /***************************************************************************************************/



    /* --- MPI RESPONSIBILITY BEGIN ---*/
    /* Initialize the driver */
    if((srioDrvHandle=Srio_Type11_init (SRIORunTimeStructures, my_cartridge_id, my_node_id))==NULL) {
    	printf("\n DJDJBG... Srio_Type11_init  failed !!!");
    	return -1;
    }
    /* --- MPI RESPONSIBILITY END ---*/

#ifdef MANUAL_RUN
    printf("\n*************  Ready ************");
    printf("\n Press any key to continue to enable PDSP..... ");
    printf("\n********************************");
    getchar();

#else
    sleep(10);
#endif

    /********************************* RUN TIME (Done by MPI)******************************************/
    /* --- MPI RESPONSIBILITY BEGIN ---*/
    printf ("\nDJDJBG... MPI Runtime :About to initialize routing structures\n"); fflush(stdout);
    /* Initialize SRIO Routing structures with the cluster information based on the cartridge_ids[]
     * from MPI or User input in this test program */
    if(form_routingTable(SRIORunTimeStructures,cartridge_ids,num_cartridge_ids)<0) {
    	printf("\n ...MPI Runtime. Routing table formation failed!!!");
    	return -1;
    }

    /* Populate the list_of_srio_ids[] */
    num_of_srio_ids=0;
    for(i=0;i<remoteInfo.num_entries;i++) {
    	list_of_srio_ids[num_of_srio_ids]=remoteInfo.entries[i].remote_id;
    	printf("\n list_of_srio_ids[%d]=%d",i,list_of_srio_ids[num_of_srio_ids]);
    	num_of_srio_ids++;
    }

    if(Srio_Type11_setup (srioDrvHandle, SRIORunTimeStructures, list_of_srio_ids,num_of_srio_ids)<0) {
      printf ("\nDJDJBG... Srio_Type11_setup failed"); fflush(stdout);
      return -1;
    }
    /* --- MPI RESPONSIBILITY END ---*/

    /*********************************TEST START ****************************************/
    if(send_receive_Test (srioDrvHandle,nodeRole,my_id, &remoteInfo)<0) {
       printf("\n Send_Receive Test failed!!\n");
    }

    if(Srio_Type11_finalize(srioDrvHandle)<0) {
        printf("\n Srio_Type11_finalize failed!!\n");
        return -1;
    }

    return 0;
}
/*-------------------------------------------------------------------------------------------*/

/*** nothing past this point ***/
