/**
 *   @file  multicoreLoopback.c
 *
 *   @brief   
 *      This is an example application which showcases how the SRIO Driver
 *      API can be used in a multiple core environment. 
 *
 *      The test suite sends/receives data as follows:-
 *
 *      CORE 1 --> CORE 2 --> CORE 3 --> CORE 0 --> CORE 1
 *
 *      At each step the received data is validated (Payload and Received
 *      Identifiers).
 *
 *      The goal of the test suite is to showcase that the SRIO Driver API
 *      are multi-core safe.
 *
 *  \par
 *  ============================================================================
 *  @n   (C) Copyright 2009-2014, Texas Instruments, Inc.
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  \par
*/
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sched.h>
#include<signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stddef.h>
#include <time.h>

/* CPPI/QMSS Include Files. */
#include <ti/drv/cppi/cppi_drv.h>
#include <ti/drv/qmss/qmss_drv.h>
#include <ti/drv/qmss/qmss_qm.h>
#include <ti/drv/qmss/qmss_srioRtr.h>
/* CSL Include Files */
#include <ti/csl/csl_chip.h>
#include <ti/csl/csl_psc.h>
#include <ti/csl/csl_pscAux.h>
#include <ti/csl/csl_srioAux.h>

#include <ti/csl/csl_srioAuxPhyLayer.h>

#include <ti/csl/csl_serdes_srio.h>

/* Device specific include */
#include "fw_test.h"
#include "fw_mem_allocator.h"
#include "linuxutil.h"
#include "srio_test.h"
#include "rtr_sriodrv.h"
#include "srioPlatCfg.h"
#include "srio_osal.h"
#include "../pktfwdK2H/mpi_interfaces.h"

/**********************************************************************
 ************************** LOCAL Definitions *************************
 **********************************************************************/





/**********************************************************************
 ************************** Global Variables **************************
 **********************************************************************/
Sriotype11APIsContext_t Sriotype11APIsContext;


/**********************************************************************
 ************************* LOCAL Definitions **************************
 **********************************************************************/



/* Memory used for the descriptors. */
Uint8   *host_region_terminate;
Uint8   *host_region_credit;
Uint8   *host_region_forward;


Qmss_InitCfg   qmssInitConfig;

Qmss_Result    qmssRegion_terminate;
Qmss_Result    qmssRegion_credit;
Qmss_Result    qmssRegion_forward;


/* Global Varialble which keeps track of the core number executing the
 * application. */

Qmss_QueueHnd           qpendRxQueHnd;
int                     qpendFd;
int                     srioFd;
Srio_DrvHandle          isr_hSrioDriver;

/**********************************************************************
 ************************* Extern Definitions *************************
 **********************************************************************/

extern uint32_t malloc_counter;
extern uint32_t free_counter;

extern int32_t SrioDevice_init (CSL_SrioHandle hSrio, uint32_t srioSerdesVAddr);
extern int32_t SrioDevice_init_lite (CSL_SrioHandle hSrio, uint32_t srioSerdesVAddr);
extern int32_t SrioDevice_deinit (CSL_SrioHandle hSrio);
extern int32_t SrioDevice_deinit_lite (CSL_SrioHandle hSrio);

extern void* Osal_srioDataBufferMalloc(uint32_t numBytes);

/* QMSS device specific configuration */
extern Qmss_GlobalConfigParams  qmssGblCfgParams;

/* OSAL Data Buffer Memory Initialization. */
extern int32_t Osal_dataBufferInitMemory(uint32_t dataBufferSize);

/**********************************************************************
 ************************* LOCAL Definitions **************************
 **********************************************************************/

/* Queue handles for garbage queues */
static Qmss_QueueHnd garbageQueueHnd[] = {
    QMSS_PARAM_NOT_SPECIFIED,
    QMSS_PARAM_NOT_SPECIFIED,
    QMSS_PARAM_NOT_SPECIFIED,
    QMSS_PARAM_NOT_SPECIFIED,
    QMSS_PARAM_NOT_SPECIFIED,
    QMSS_PARAM_NOT_SPECIFIED
};

/* These are the GARBAGE queues which are used by the TXU to dump the
 * descriptor if there is an error instead of recycling the descriptor
 * to the free queue. */
#define GARBAGE_LEN_QUEUE           905
#define GARBAGE_TOUT_QUEUE          906
#define GARBAGE_RETRY_QUEUE         907
#define GARBAGE_TRANS_ERR_QUEUE     908
#define GARBAGE_PROG_QUEUE          909
#define GARBAGE_SSIZE_QUEUE         910

/* SRIO Device Information
 * - 16 bit Device Identifier.
 * - 8 bit Device Identifier.
 * - Vendor Identifier.
 * - Device Revision. */
#define DEVICE_VENDOR_ID            0x30
#define DEVICE_REVISION             0x0

/* SRIO Assembly Information
 * - Assembly Identifier
 * - Assembly Vendor Identifier.
 * - Assembly Device Revision.
 * - Assembly Extension Features */
#define DEVICE_ASSEMBLY_ID          0x0
#define DEVICE_ASSEMBLY_VENDOR_ID   0x30
#define DEVICE_ASSEMBLY_REVISION    0x0
#define DEVICE_ASSEMBLY_INFO        0x0100

/**************************************************************************************************************************************************/
/**************************************************************************************************************************************************/
/**************************************************************************************************************************************************/
/**************************************************************************************************************************************************/

static Int32 system_init (void)
{
    Qmss_MemRegInfo     memRegInfo;

   memset (&memRegInfo, 0, sizeof(memRegInfo));
    /* Memory Region 0 Configuration */
    memRegInfo.descBase         = (uint32_t *)host_region_terminate;
    memRegInfo.descSize         = SIZE_HOST_DESC;
    memRegInfo.descNum          = NUM_TERMINATE_PACKETS_DDR;
    memRegInfo.manageDescFlag   = Qmss_ManageDesc_MANAGE_DESCRIPTOR;
    memRegInfo.memRegion        = Qmss_MemRegion_MEMORY_REGION_NOT_SPECIFIED;
    memRegInfo.startIndex       = QMSS_START_INDEX_INTERNAL;

    /* Initialize and inset the memory region. */
    qmssRegion_terminate = Qmss_insertMemoryRegion (&memRegInfo);
    if (qmssRegion_terminate < QMSS_SOK)
    {
        printf ("Error inserting host terminate memory region: %d\n",  qmssRegion_terminate);
        return -1;
    }



    /* Share the region */
//    ex_rm_name_set ("srio_mt_region", qmssRegion_terminate);

    /* CPPI and Queue Manager are initialized. */
    Srio_osalLog ("Debug: Queue Manager and CPPI for TERMINATE   are initialized.\n");
    Srio_osalLog ("Debug: Host Region for TERMINATE %p\n", host_region_terminate);

     /********************************  CREDIT  *************************************/

        /* QMSS Region for Credit buffers */
        memset (&memRegInfo, 0, sizeof(memRegInfo));
        /* Memory Region 0 Configuration */
        memRegInfo.descBase         = (uint32_t *)host_region_credit;
        memRegInfo.descSize         = SIZE_HOST_DESC;
        memRegInfo.descNum          = NUM_CREDIT_PACKETS_MSMC;
        memRegInfo.manageDescFlag   = Qmss_ManageDesc_MANAGE_DESCRIPTOR;
        memRegInfo.memRegion        = Qmss_MemRegion_MEMORY_REGION_NOT_SPECIFIED;
        memRegInfo.startIndex       = QMSS_START_INDEX_INTERNAL;

        /* Initialize and inset the memory region. */
        qmssRegion_credit = Qmss_insertMemoryRegion (&memRegInfo);
        if (qmssRegion_credit < QMSS_SOK)
        {
        	printf ("Error inserting credit_msmc memory region: %d\n", qmssRegion_credit);
            return -1;
        }



        /* Share the region */
 //       ex_rm_name_set ("srio_mt_region_credit", qmssRegion_credit);

        /* CPPI and Queue Manager are initialized. */
        Srio_osalLog ("Debug: Queue Manager and CPPI for CREDIT are initialized.\n");
        Srio_osalLog ("Debug: Host Region for CREDIT  %p\n", host_region_credit);

        /********************************  FORWARD  *************************************/

        /* QMSS Region for Credit buffers */
        memset (&memRegInfo, 0, sizeof(memRegInfo));
        /* Memory Region 0 Configuration */
        memRegInfo.descBase         = (uint32_t *)host_region_forward;
        memRegInfo.descSize         = SIZE_HOST_DESC;
        memRegInfo.descNum          = NUM_FWD_PACKETS_MSMC;// NUM_FWD_PACKETS_MSMC Aligned to 2^ minimum
        memRegInfo.manageDescFlag   = Qmss_ManageDesc_MANAGE_DESCRIPTOR;
        memRegInfo.memRegion        = Qmss_MemRegion_MEMORY_REGION_NOT_SPECIFIED;
        memRegInfo.startIndex       = QMSS_START_INDEX_INTERNAL;

        /* Initialize and inset the memory region. */
        qmssRegion_forward = Qmss_insertMemoryRegion (&memRegInfo);
        if (qmssRegion_forward < QMSS_SOK)
        {
        	printf ("Error inserting forward memory region: %d\n",qmssRegion_forward);
              return -1;
        }

        /* Share the region */
//        ex_rm_name_set ("srio_mt_region_credit", qmssRegion_forward);

        /* CPPI and Queue Manager are initialized. */
        Srio_osalLog ("Debug: Queue Manager and CPPI for CREDIT are initialized.\n");
        Srio_osalLog ("Debug: Host Region for CREDIT  %p\n",host_region_forward);


    return 0;
}


/**
 *  @b Description
 *  @n
 *      System Initialization Code. This is added here only for illustrative
 *      purposes and needs to be invoked once during initialization at
 *      system startup.
 *
 *  @retval
 *      Success     -   0
 *  @retval
 *      Error       -   <0
 */
/**************************************************************************************************************************************************/
/**************************************************************************************************************************************************/
static Int32 system_deinit (void)
{
    Qmss_Result         qmssResult;
    Cppi_Result         cppiResult;


    /* remove the memory region. */
    qmssResult = Qmss_removeMemoryRegion (qmssRegion_terminate, 0);
    if (qmssResult < QMSS_SOK)
    {
    	printf (" Error removing terminate memory region: %d\n",  qmssResult);
        return -1;
    }

    qmssResult = Qmss_removeMemoryRegion (qmssRegion_credit, 0);
    if (qmssResult < QMSS_SOK)
    {
    	printf (" Error removing credit memory region: %d\n",  qmssResult);
        return -1;
    }


    qmssResult = Qmss_removeMemoryRegion (qmssRegion_forward, 0);
        if (qmssResult < QMSS_SOK)
        {
        	printf (" Error removing forward memory region: %d\n",  qmssResult);
            return -1;
        }


    /* De Initialize CPPI CPDMA */
    cppiResult = Cppi_exit ();
    if (cppiResult != CPPI_SOK)
    {
    	printf ("Error deinitializing CPPI error code : %d\n",  cppiResult);
        return -1;
    }

    return 0;
}

/**********************************************************************
 ************************ SRIO TEST FUNCTIONS *************************
 **********************************************************************/
/**
 *  @b Description
 *  @n
 *      Yields the task using something like sleep() or usleep().
 *
 *  @retval
 *      Not Applicable
 */
void yield (void)
{
    sched_yield();
}

/**
 *  @b Description
 *  @n
 *      Utility to feed a command to rm
 *
 *  @param[in]  name
 *      Name to query
 *  @param[in]  val
 *      return value or value to set
 *  @param[in]  type
 *      transaction type
 *  @param[in]  name
 *      retry until success
 *
 *  @retval
 *      Value
 */
void ex_rm_cmd (const char *name, uint32_t *val, Rm_ServiceType type, int retry)
{
    Rm_ServiceReqInfo   rmServiceReq;
    Rm_ServiceRespInfo  rmServiceResp;
    int                 succeed;

    memset((void *)&rmServiceReq, 0, sizeof(rmServiceReq));
    memset((void *)&rmServiceResp, 0, sizeof(rmServiceResp));

    rmServiceReq.type             = type;
    rmServiceReq.resourceName     = name;
    rmServiceReq.resourceNsName   = name;
    rmServiceReq.resourceLength   = 1;
    if (val)
    {
        rmServiceReq.resourceBase = *val;
    }

    /* RM will block until resource is returned since callback is NULL */
    rmServiceReq.callback.serviceCallback = NULL;
    do {
        rmClientServiceHandle->Rm_serviceHandler(rmClientServiceHandle->rmHandle, &rmServiceReq, &rmServiceResp);
        succeed = (rmServiceResp.serviceState == RM_SERVICE_APPROVED) ||
                  (rmServiceResp.serviceState == RM_SERVICE_APPROVED_STATIC);
        if (retry && (! succeed))
        {
            yield();
        }
    } while (retry && (! succeed));

    if (succeed)
    {
        if ((type == Rm_service_RESOURCE_GET_BY_NAME) && (val))
        {
            *val = rmServiceResp.resourceBase;
        }
    }
    else
    {
    	printf ("Core %d: failed rm transaction %d %s %d (%d)\n", coreNum, type, name, val ? *val : 0, rmServiceResp.serviceState);
        errorCount++;
    }
}

/**
 *  @b Description
 *  @n
 *      Utility to look up a name in RM
 *
 *  @param[in]  name
 *      Name to query
 *
 *  @retval
 *      Value
 */
uint32_t ex_rm_name_lookup (const char *name)
{
    uint32_t val;
    ex_rm_cmd (name, &val, Rm_service_RESOURCE_GET_BY_NAME, 1);

    return val;
}

/**
 *  @b Description
 *  @n
 *      Utility to set a name in RM
 *
 *  @param[in]  name
 *      Name to query
 *
 *  @param[in]  val
 *      Value
 *
 *  @retval
 *      None
 */
void ex_rm_name_set (const char *name, uint32_t val)
{
    ex_rm_cmd (name, &val, Rm_service_RESOURCE_MAP_TO_NAME, 0);
}

/**
 *  @b Description
 *  @n
 *      Utility to delete a name from RM
 *
 *  @param[in]  name
 *      Name to delete
 *
 *  @retval
 *      None
 */
void ex_rm_name_del (const char *name)
{
    ex_rm_cmd (name, NULL, Rm_service_RESOURCE_UNMAP_NAME, 0);
}

/**
 *  @b Description
 *  @n
 *      This function enables the power/clock domains for SRIO.
 *
 *  @retval
 *      Not Applicable.
 */



/**
 *  @b Description
 *  @n
 *      The function is main run-time function for node-to-node communication
 *
 *  @param[in]  hSrioDrv
 *      Handle to the SRIO driver
 *
 *  @retval
 *      Success     - 0
 *  @retval
 *      Error       - <0
 */
/* PDSP RTR Changes */
/* Check if valid queue handle */
void check_result (const char *desc, Qmss_Result hnd)
{
    if (hnd < QMSS_SOK)
    {
    	printf("Failed to allocate queue (%d): %s\n", hnd, desc);
        exit(1);
    }
}

/* check if two memories of uint32_t are the same */
void check_memory (const char *desc, const void *buf1, const void *buf2, uint32_t size)
{
    if (memcmp (buf1, buf2, size) != 0)
    {
    	printf("Buffer compare failed: %s\n", desc);
        exit(1);
    }
}

/* Check for NULL pointer */
void check_null (const char *desc, const void *ptr)
{
    if (! ptr)
    {
    	printf("NULL check failed: %s\n", desc);
        exit(1);
    }
}
/*******************************************************************************************************/
/*******************************************************************************************************/

void check_stats (int port, Qmss_SrioRtrStats *check)
{
    Qmss_SrioRtrStats stats;
    check_result ("query stats", Qmss_getSrioRtrStats (TEST_PDSP, port, 0, 0, &stats));
    if (stats.hostPkts != check->hostPkts) {
        Srio_osalLog ("port %d stats: expect %d host packets, found %d\n", port, check->hostPkts, stats.hostPkts);
	errorCount++;
    }
    if (stats.packetsRx != check->packetsRx) {
        Srio_osalLog ("port %d stats: expect %d rx packets, found %d\n", port, check->packetsRx, stats.packetsRx);
	errorCount++;
    }
    if (stats.packetsTx != check->packetsTx) {
        Srio_osalLog ("port %d stats: expect %d tx packets, found %d\n", port, check->packetsTx, stats.packetsTx);
	errorCount++;
    }
    if (stats.creditsTx != check->creditsTx) {
        Srio_osalLog ("port %d stats: expect %d tx credits packets, found %d\n", port, check->creditsTx, stats.creditsTx);
	errorCount++;
    }
    if (stats.creditsRx != check->creditsRx) {
        Srio_osalLog ("port %d stats: expect %d rx credits packets, found %d\n", port, check->creditsRx, stats.creditsRx);
	errorCount++;
    }
}
/*******************************************************************************************************/
/*******************************************************************************************************/

/********************************  Set up queues needed by the firmware **************************/
void assignTxSRIOQueuesToHW(CSL_SrioHandle hSrio, SrioQueues_t *srioQueues){

    /**************************************************/
    /* Queues are allocated through QmssOpen(QMSS_SRIO_QUEUE). */
    /**************************************************/

      int port;
      Uint8           txQueue,prio_crf;

      /* Map the tx_srio_queues[] */
      for(port=0;port<4;port++) {

    	  txQueue = srioQueues->tx_srio_queues[port]-SRIO_QUEUE_BASE;
    	  prio_crf=SRIO_TX_DATA_QUEUE_PRIORITY<<1;/* 4 bit field,bits 0 - Crf, bit 1-2 priority,  bit 3 - Rsvd (0) */
    	  CSL_SRIO_SetTxQueueSchedInfo(hSrio, txQueue, port, prio_crf);
          Srio_osalLog("\nassignTxSRIOQueuesToHW: Mapping tx_srio_queues[%d]= %d, i.e txQueue=%d with prio_crf= %d ",port,srioQueues->tx_srio_queues[port],txQueue,prio_crf);
          /* Map the tx_credit_srio_queues[] */
    	  txQueue = srioQueues->tx_srio_credit_queues[port]-SRIO_QUEUE_BASE;
    	  prio_crf=SRIO_TX_CREDIT_QUEUE_PRIORITY<<1 | 0x0;/* 4 bit field,bits 0 - Crf, bit 1-2 priority,  bit 3 - Rsvd (0) */
    	                                                  /* For Exp.1  Credit packets's CRF=0 as well */
    	  CSL_SRIO_SetTxQueueSchedInfo(hSrio, txQueue, port, prio_crf);
    	  Srio_osalLog("\nassignTxSRIOQueuesToHW: Mapping tx_srio_credit_queues[%d]= %d, i.e txQueue=%d with prio_crf= %d ",port,srioQueues->tx_srio_credit_queues[port],txQueue,prio_crf);
      }
        Srio_osalLog("\n");
}
#ifdef BYPASS_PDSP
Qmss_QueueHnd saved_tx_terminate_queues[4];
#endif
/********************************  Set up queues needed by the firmware lie**************************/
int allocate_SRIOQueues(SrioQueues_t *srioQueues,int dest_id[]) {

    int                     port;
    Qmss_QueueHnd           fwQHnds[QMSS_SRIORTR_FW_QUEUES];
    Qmss_QueueHnd baseQueue;
    uint8_t isAllocated;

    /* Allocate block of queues to be used by firmware */
    baseQueue = Qmss_queueBlockOpen (fwQHnds, Qmss_QueueType_GENERAL_PURPOSE_QUEUE, QMSS_SRIORTR_FW_QUEUES, 32);
    if (baseQueue < 0)
    {
    	printf ("Failed to open %d contiguous queues\n",QMSS_SRIORTR_FW_QUEUES);
        errorCount++;
    }
    srioQueues->queueBase=baseQueue;

    /********************************  Set up queues needed by the firmware **************************/
    /* Allocate block of queues to be used by firmware for host,forwarding, and credit. 3 categories, 4 ports per category.
     * QMSS_SRIORTR_INPUT_QUEUEOFFSET_HOST_QUEUE(port), QMSS_SRIORTR_INPUT_QUEUEOFFSET_FWD_QUEUE(port), and
     * QMSS_SRIORTR_INPUT_QUEUEOFFSET_CREDIT_QUEUE(port) provides the queue number within this array of queues,i.e fwQHnds[] .
     *  */
    /* Opening Tx Termination Queues */
    for(port=0;port<4;port++) {
    	/************ Terminiation Queues ******************/

    	srioQueues->tx_terminate_queues[port]=fwQHnds[QMSS_SRIORTR_INPUT_QUEUEOFFSET_HOST_QUEUE(port)];
#ifdef BYPASS_PDSP
        saved_tx_terminate_queues[port]=srioQueues->tx_terminate_queues[port]; /* Saving to close   */
#endif
    	srioQueues->txFree_terminate_queues[port]=Qmss_queueOpen (Qmss_QueueType_GENERAL_PURPOSE_QUEUE, QMSS_PARAM_NOT_SPECIFIED, &isAllocated);
        if (srioQueues->txFree_terminate_queues[port] < 0)
        {
         	printf ("Error: Tx Free Terminate Queue failed to open\n");
             return -1;
        }

        /************ Credit Queues ******************/
        srioQueues->tx_srio_credit_queues[port]=Qmss_queueOpen (Qmss_QueueType_SRIO_QUEUE, QMSS_PARAM_NOT_SPECIFIED, &isAllocated);
        if (srioQueues->tx_srio_credit_queues[port] < 0)
        {
             	printf ("Error: Tx Credit Queue failed to open\n");
            return -1;
        }

            srioQueues->txFree_credit_queues[port]=Qmss_queueOpen (Qmss_QueueType_GENERAL_PURPOSE_QUEUE, QMSS_PARAM_NOT_SPECIFIED, &isAllocated);
            if (srioQueues->txFree_credit_queues[port] < 0)
            {
              	printf ("Error: Tx Free Credit Queue failed to open\n");
                  return -1;
            }


            srioQueues->rx_credit_queues[port]=fwQHnds[QMSS_SRIORTR_INPUT_QUEUEOFFSET_CREDIT_QUEUE(port)];
            srioQueues->rxFree_credit_queues[port]=Qmss_queueOpen (Qmss_QueueType_GENERAL_PURPOSE_QUEUE, QMSS_PARAM_NOT_SPECIFIED, &isAllocated);
            if (srioQueues->rxFree_credit_queues[port] < 0)
            {
             	printf ("Error: Rx Free Credit Queue failed to open\n");
                 return -1;
            }

            /************ TX SRIO HW Queues ******************/
            srioQueues->tx_srio_queues[port]=Qmss_queueOpen (Qmss_QueueType_SRIO_QUEUE, QMSS_PARAM_NOT_SPECIFIED, &isAllocated);
            if (srioQueues->tx_srio_queues[port] < 0)
            {
             	printf ("Error: Tx SRIO HW Queue failed to open\n");
                 return -1;
            }

#ifdef BYPASS_PDSP
          srioQueues->tx_terminate_queues[port]=srioQueues->tx_srio_queues[port];
#endif

             /************ RX SRIO HW Queues ******************/
            srioQueues->rx_completed_forward_queues[port]=fwQHnds[QMSS_SRIORTR_INPUT_QUEUEOFFSET_FWD_QUEUE(port)];
            srioQueues->rx_forward_return_queues[port]=fwQHnds[QMSS_SRIORTR_INPUT_QUEUEOFFSET_FWD_RET_QUEUE(port)];

            srioQueues->rxFree_forward_queues[port]=Qmss_queueOpen (Qmss_QueueType_GENERAL_PURPOSE_QUEUE, QMSS_PARAM_NOT_SPECIFIED, &isAllocated);
             if (srioQueues->rxFree_forward_queues[port] < 0)
             {
              	printf ("Error: Rx Free rxFree_forward_queues failed to open\n");
                  return -1;
             }

             /************ Debug Queues ******************/
            srioQueues->debugQueues[port]=Qmss_queueOpen (Qmss_QueueType_GENERAL_PURPOSE_QUEUE, QMSS_PARAM_NOT_SPECIFIED, &isAllocated);
    }

    /* Thereis only One Rx Termination Queue, not per port */
    srioQueues->rx_terminate_queue=Qmss_queueOpen (Qmss_QueueType_GENERAL_PURPOSE_QUEUE, QMSS_PARAM_NOT_SPECIFIED, &isAllocated);
    if (srioQueues->rx_terminate_queue < 0)
    {
     	printf ("Error: Rx Terminate Queue failed to open\n");
         return -1;
    }

    srioQueues->rxFree_terminate_queue=Qmss_queueOpen (Qmss_QueueType_GENERAL_PURPOSE_QUEUE, QMSS_PARAM_NOT_SPECIFIED, &isAllocated);
    if (srioQueues->rxFree_terminate_queue < 0)
    {
     	printf ("Error: Rx Free Terminate Queue failed to open\n");
         return -1;
    }

   return 0;
}
/********************************  Set up queues needed by the firmware lie**************************/
/*******************************************************************************************************/
/*******************************************************************************************************/
int display_SRIOQueues(SrioQueues_t *srioQueues) {

    int                     port;

    /********************************  Set up queues needed by the firmware **************************/
    /* Allocate block of queues to be used by firmware for host,forwarding, and credit. 3 categories, 4 ports per category.
     * QMSS_SRIORTR_INPUT_QUEUEOFFSET_HOST_QUEUE(port), QMSS_SRIORTR_INPUT_QUEUEOFFSET_FWD_QUEUE(port), and
     * QMSS_SRIORTR_INPUT_QUEUEOFFSET_CREDIT_QUEUE(port) provides the queue number within this array of queues,i.e fwQHnds[] .
     *  */
    /* Opening Tx Termination Queues */
    Srio_osalLog("\n\n########################### ALLOCATED QUEUES  #######################\n");

    for(port=0;port<4;port++) {
    	/************ Terminiation Queues ******************/
        Srio_osalLog("\n**********************************************************************\n");
        Srio_osalLog("\n -------- SRIO HW interfacing Queues(Port %d)-------------------\n",port);
        Srio_osalLog("\n tx_terminate_queues[%d]         = %d",port,Qmss_getQIDFromHandle(srioQueues->tx_terminate_queues[port]));
        Srio_osalLog("\n tx_srio_queues[%d]              = %d",port,Qmss_getQIDFromHandle(srioQueues->tx_srio_queues[port]));
        Srio_osalLog("\n tx_srio_credit_queues[%d]       = %d",port,Qmss_getQIDFromHandle(srioQueues->tx_srio_credit_queues[port]));

        Srio_osalLog("\n -------------------SRIO RX Queues (Port %d) ----------------..\n",port);
        Srio_osalLog("\n rx_credit_queues[%d]            = %d",port,Qmss_getQIDFromHandle(srioQueues->rx_credit_queues[port]));
        Srio_osalLog("\n rx_completed_forward_queues[%d] = %d",port,Qmss_getQIDFromHandle(srioQueues->rx_completed_forward_queues[port]));

        Srio_osalLog("\n --------------------- Free Queues (Port %d) ----------------------\n",port);
        Srio_osalLog("\n txFree_terminate_queues[%d]     = %d",port,Qmss_getQIDFromHandle(srioQueues->txFree_terminate_queues[port]));
        Srio_osalLog("\n txFree_credit_queues[%d]        = %d",port,Qmss_getQIDFromHandle(srioQueues->txFree_credit_queues[port]));
        Srio_osalLog("\n rxFree_credit_queues[%d]        = %d",port,Qmss_getQIDFromHandle(srioQueues->rxFree_credit_queues[port]));
        Srio_osalLog("\n rxFree_forward_queues[%d]       = %d",port,Qmss_getQIDFromHandle(srioQueues->rxFree_forward_queues[port]));

        Srio_osalLog("\n**********************************************************************\n");
    }
    Srio_osalLog("\n queueBase=             = %d",Qmss_getQIDFromHandle(srioQueues->queueBase));
    Srio_osalLog("\n rx_terminate_queue     = %d",Qmss_getQIDFromHandle(srioQueues->rx_terminate_queue));
	printf("\n rxFree_terminate_queue = %d",Qmss_getQIDFromHandle(srioQueues->rxFree_terminate_queue));

	printf("\n\n######################## ALLOCATED QUEUES END ########################\n\n");

  return 0;
}

/******************************************** Close all the queues **********************************/
/************************************************************************************************/
int deallocate_queue(Qmss_QueueHnd queueHnd,int numBuffers,int mtu_size) {

  int idx;
  Cppi_Desc * ptrHostDesc;
  uint8_t            *ptr_dataBuffer;
  Qmss_Result qmssResult;
  uint32_t            dataBufferLen;

  for (idx = 0; idx < numBuffers; idx++)
  {
     /* Get a host descriptor from the free TX queue. */
     ptrHostDesc = (Cppi_Desc *)QMSS_DESC_PTR(Qmss_queuePop(queueHnd));
     if (ptrHostDesc == NULL)
     {
         break;
     }
     /* Extract the buffer pointer */
     Cppi_getData (Cppi_DescType_HOST, ptrHostDesc, &ptr_dataBuffer, &dataBufferLen);
     /* Free the buffer */
     Srio_osalDataBufferFree (ptr_dataBuffer,mtu_size);
     /* Intentionally discard the descriptor */
   }

   /* Close the SRIO Transmit Free Queue; this should support starvation. */
   if ((qmssResult = Qmss_queueClose (queueHnd)) < QMSS_SOK)
   {
     printf ("Error: SRIO Transmit Completion Queue failed to close: %d\n", qmssResult);
     return -1;
   }
   return(idx);
}
/*******************************************************************************************************/
void djdbg_report_psinfo (void *desc)
{
uint32_t *srioInfo;
uint32_t  psLen;
uint16_t  dstID, srcID;
uint8_t   *bufAddr,mbox, letter, ccType11, priType11, ttType11;
uint32_t buffLen;

     Cppi_getPSData (Cppi_getDescType(desc), Cppi_PSLoc_PS_IN_DESC, desc, (uint8_t**)&srioInfo, &psLen);
     Cppi_getData (Cppi_getDescType(desc), desc, &bufAddr, &buffLen);
     CSL_SRIO_ExtractType11PSInfo(srioInfo, &dstID, &srcID, &mbox, &letter, &ttType11, &priType11, &ccType11);
     Srio_osalLogC ("XwrongX psLen=%d dstID=%02x srcID=%02x mbox=%02x letter=%02x ttType11=%02x pri=%02x ccType11=%02x buffLen=%d RAW[%08x %08x]",
                   psLen, dstID, srcID, mbox, letter, ttType11, priType11, ccType11,buffLen,
                   srioInfo[0], srioInfo[1]); 
}
/*******************************************************************************************************/

int clean_garbage ()
{
   void *desc;
   int total_desc = 0,i;
   uint32_t *desc_ptr;
   int queue;
   /* DJDBG: Debug vars: */
   uint32_t *srioInfo;
   uint32_t  psLen;
   uint16_t  dstID, srcID;
   uint8_t   *bufAddr,mbox, letter, ccType11, priType11, ttType11;
   uint32_t buffLen;
   int ret;

   for (queue = 905; queue <= 910; queue++)
   {
	   int desc_found = 0;
	   while ( desc = (Cppi_Desc *)QMSS_DESC_PTR(Qmss_queuePop((Qmss_QueueHnd)queue)))
	   {
              desc_found++;
              /* DJDBG: */
              Cppi_getPSData (Cppi_getDescType(desc), Cppi_PSLoc_PS_IN_DESC, desc, (uint8_t**)&srioInfo, &psLen);
              Cppi_getData (Cppi_getDescType(desc), desc, &bufAddr, &buffLen);
              CSL_SRIO_ExtractType11PSInfo(srioInfo, &dstID, &srcID, &mbox, &letter, &ttType11, &priType11, &ccType11);
              Srio_osalLog ("\n[%2d] psLen=%d dstID=%02x srcID=%02x mbox=%02x letter=%02x ttType11=%02x pri=%02x ccType11=%02x buffLen=%d", desc_found,
                psLen, dstID, srcID, mbox, letter, ttType11, priType11, ccType11,buffLen); fflush(stdout);

              Srio_osalLog("\n Raw data: ");
              Srio_osalLog("\n CPPI DESC: ");
              desc_ptr=(uint32_t *)desc;
              for(i=0;i<8;i++) {
                  Srio_osalLog(" %08x",desc_ptr[i]);
	          }
              Srio_osalLog("\n PS INFO : ");
              for(i=8;i<12;i++) {
                  Srio_osalLog(" %08x",desc_ptr[i]);
	          }

	   }
	   if (desc_found)
	   {
	      total_desc += desc_found;
	      Srio_osalLog("\nFound GARBAGE[%d] = %d\n", queue, desc_found);
	   }

	   ret=deallocate_queue(queue,desc_found,psLen);
	   Srio_osalLog("\n Freed %d descriptors",ret);
   }

   return total_desc;
}

/*******************************************************************************************************/
void clean_one_queue (int queue)
{
   void *desc;
   uint32_t *desc_ptr;
   uint32_t *srioInfo;
   uint32_t  psLen;
   uint16_t  dstID, srcID;
   uint8_t   *bufAddr,mbox, letter, ccType11, priType11, ttType11;
   uint32_t buffLen;
   int ret;
   int desc_found = 0;

    while ( desc = (Cppi_Desc *)QMSS_DESC_PTR(Qmss_queuePop((Qmss_QueueHnd)queue)))
    {
      desc_found++;
      Cppi_getPSData (Cppi_getDescType(desc), Cppi_PSLoc_PS_IN_DESC, desc, (uint8_t**)&srioInfo, &psLen);
      Cppi_getData (Cppi_getDescType(desc), desc, &bufAddr, &buffLen);
      CSL_SRIO_ExtractType11PSInfo(srioInfo, &dstID, &srcID, &mbox, &letter, &ttType11, &priType11, &ccType11);
      Srio_osalLog ("CLEAN [%3d] psLen=%d dstID=%02x srcID=%02x mbox=%02x letter=%02x ttType11=%02x pri=%02x ccType11=%02x buffLen=%d", desc_found,
                psLen, dstID, srcID, mbox, letter, ttType11, priType11, ccType11,buffLen); fflush(stdout);

      desc_ptr=(uint32_t *)desc;
      //for(i=0;i<8;i++) Srio_osalLog(" %08x",desc_ptr[i]);
      Srio_osalLog("CLEAN SRIO PS INFO : %08x %08x %08x %08x ", desc_ptr[8], desc_ptr[9], desc_ptr[10], desc_ptr[11]);
    }
    if(desc_found) {
      //ret=deallocate_queue(queue,desc_found,psLen);
      //Srio_osalLog("CLEAN Freed %d descriptors", ret);
    }
}
/*******************************************************************************************************/

int clean_debug_queues ()
{
   void *desc;
   int total_desc = 0,i;
   int queue;
   /* DJDBG: Debug vars: */
   uint32_t *srioInfo;
   uint32_t  psLen;
   uint16_t  dstID, srcID;
   uint8_t   mbox, letter, ccType11, priType11, ttType11;
   int ret,port;
   uint32_t *desc_ptr;


   for(port=0;port<4;port++) {
	  queue = Sriotype11APIsContext.srioQueues.debugQueues[port];

	   int desc_found = 0;
	   while ( desc = (Cppi_Desc *)QMSS_DESC_PTR(Qmss_queuePop((Qmss_QueueHnd)queue)))
	   {
              desc_found++;
              /* DJDBG: */
              Cppi_getPSData (Cppi_getDescType(desc), Cppi_PSLoc_PS_IN_DESC, desc, (uint8_t**)&srioInfo, &psLen);
              CSL_SRIO_ExtractType11PSInfo(srioInfo, &dstID, &srcID, &mbox, &letter, &ttType11, &priType11, &ccType11);
              Srio_osalLog ("\n DBGQUEUE[Port %d] [%2d] psLen=%d dstID=%02x srcID=%02x mbox=%02x letter=%02x ttType11=%02x pri=%02x ccType11=%02x", port,desc_found,
                psLen, dstID, srcID, mbox, letter, ttType11, priType11, ccType11); fflush(stdout);

              Srio_osalLog("\n Raw data: ");
              desc_ptr=(uint32_t *)desc;
              for(i=8;i<12;i++) {
                  Srio_osalLog(" %08x",desc_ptr[i]);
	          }

	   }
	   if (desc_found)
	   {
	      total_desc += desc_found;
	      Srio_osalLog("\nFound in DebugQueue[%d] = %d\n", queue, desc_found);
	   }

	   ret=deallocate_queue(queue,desc_found,psLen);
	   Srio_osalLog("\n Freed %d descriptors",ret);
   }

   return total_desc;
}

/******************************************** Close all the queues **********************************/
/************************************************************************************************/
int deallocate_SRIOQueues(SrioQueues_t *srioQueues) {

    int                     port;
    Qmss_QueueHnd           fwQHnds[QMSS_SRIORTR_FW_QUEUES];
    Qmss_QueueHnd baseQueue;
    uint8_t isAllocated;
    int count=0,ret,total_descriptors_allocated;
    Srio_osalLog("\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ DEALLOCATION ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n");
     if((ret=deallocate_queue(srioQueues->rx_terminate_queue,NUM_TERMINATE_RX_DESCS,SRIO_MAX_MTU))<0) {
        printf("\nRx Terminate Queues Failed ");
    } else {
    	count+=ret;
        Srio_osalLog("\nRx_terminate_queue     (Common)       SUCESSS    deallocated=%d, count so far= %d",ret,count);
    }

    if((ret=deallocate_queue(srioQueues->rxFree_terminate_queue,NUM_TERMINATE_RX_DESCS,SRIO_MAX_MTU))<0) {
    	printf("\n Deallocating RxFree Terminate Queues Failed ");
    } else {
    	count+=ret;
        Srio_osalLog("\nRxFree_Terminate_queue (Common)       SUCCESS      deallocated=%d, count so far= %d",ret,count);
    }

    Srio_osalLog("\n\n");
   for(port=0;port<4;port++) {

#ifdef BYPASS_PDSP
	   srioQueues->tx_terminate_queues[port]=saved_tx_terminate_queues[port];
#endif

	   if((ret=deallocate_queue(srioQueues->tx_terminate_queues[port],NUM_TERMINATE_TX_DESCS/4,SRIO_MAX_MTU))<0) {
		   printf("\n Deallocating Terminate Queues Failed for port %d",port);
    } else {
    	count+=ret;
        Srio_osalLog("\ntxTerminate Queues             SUCCESS for port %d, deallocated=%d, count so far= %d",port,ret,count);
    }
    if((ret=deallocate_queue(srioQueues->txFree_terminate_queues[port],NUM_TERMINATE_TX_DESCS/4,SRIO_MAX_MTU))<0) {
    	printf("\n Deallocating TxFree Terminate Queues Failed for port %d",port);
    } else {
    	count+=ret;
        Srio_osalLog("\ntxFree_terminate_queues      SUCCESS for port %d, deallocated=%d, count so far= %d",port,ret,count);
    }
    Srio_osalLog("\n");
     /* Credit Queues */
    if((ret=deallocate_queue(srioQueues->tx_srio_credit_queues[port],NUM_CREDIT_TX_DESCS/4,SRIO_MAX_CREDIT_MTU))<0) {
    	printf("\ntx_srio_credit_queues Failed for port %d",port);
    } else {
    	count+=ret;
        Srio_osalLog("\ntx_srio_credit_queues(HW)    SUCCESS for port %d, deallocated=%d, count so far= %d",port,ret,count);
    }

    if((ret=deallocate_queue(srioQueues->txFree_credit_queues[port],NUM_CREDIT_TX_DESCS/4,SRIO_MAX_CREDIT_MTU))<0) {
    	printf("\ntxFree_credit_queues Failed for port %d",port);
    } else {
    	count+=ret;
        Srio_osalLog("\ntxFree_credit_queues         SUCCESS for port %d, deallocated=%d, count so far= %d",port,ret,count);
    }
    Srio_osalLog("\n");
    /* TX SRIO Queues */
    if((ret=deallocate_queue(srioQueues->tx_srio_queues[port],NUM_CREDIT_RX_DESCS/4,SRIO_MAX_MTU))<0) {
    	printf("\n Deallocating tx_srio_queues Failed for port %d",port);
    } else {
    	count+=ret;
       Srio_osalLog("\ntx_srio_queues(HW)           SUCCESS for port %d, deallocated=%d, count so far= %d",port,ret,count);
    }
    /* Forward Queues */

    Srio_osalLog("\n");
    if((ret=deallocate_queue(srioQueues->rx_credit_queues[port],NUM_CREDIT_RX_DESCS/4,SRIO_MAX_CREDIT_MTU))<0) {
    	printf("\nrx_credit_queues Failed for port %d",port);
     } else {
     	count+=ret;
        Srio_osalLog("\n rx_credit_queues             SUCCESS for port %d, deallocated=%d, count so far= %d",port,ret,count);
     }

     if((ret=deallocate_queue(srioQueues->rxFree_credit_queues[port],NUM_CREDIT_RX_DESCS/4,SRIO_MAX_CREDIT_MTU))<0) {
    	 printf("\n Deallocating rxFree_credit_queues Failed for port %d",port);
     } else {
     	count+=ret;
        Srio_osalLog("\nrxFree_credit_queues         SUCCESS for port %d, deallocated=%d, count so far= %d",port,ret,count);
     }
     Srio_osalLog("\n");
     if((ret=deallocate_queue(srioQueues->rx_completed_forward_queues[port],NUM_FWD_POOL_PER_PORT,SRIO_MAX_MTU))<0) {
    	 printf("\n Deallocating rx_completed_forward_queues Failed for port %d",port);
     } else {
     	count+=ret;
        Srio_osalLog("\nrx_completed_forward_queues  SUCCESS for port %d, deallocated=%d, count so far= %d",port,ret,count);
     }

     if((ret=deallocate_queue(srioQueues->rxFree_forward_queues[port],NUM_FWD_POOL_PER_PORT,SRIO_MAX_MTU))<0) {
    	 printf("\n Deallocating rxFree_forward_queues Failed for port %d",port);
     } else {
     	count+=ret;
        Srio_osalLog("\nrxFree_forward_queues        SUCCESS for port %d, deallocated=%d, count so far= %d",port,ret,count);
     }

     if((ret=deallocate_queue(srioQueues->rx_forward_return_queues[port],NUM_FWD_POOL_PER_PORT,SRIO_MAX_MTU))<0) {
    	 printf("\n Deallocating rx_forward_return_queues Failed for port %d",port);
     } else {
     	count+=ret;
        Srio_osalLog("\nrx_forward_return_queues        SUCCESS for port %d, deallocated=%d, count so far= %d",port,ret,count);
     }

     Srio_osalLog("\n\n\n");
   }
   Srio_osalLog("\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ DEALLOCATION of QUEUES END ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n");

   Srio_osalLog("\n Deallocating Garbage Queues ");
    count += clean_garbage();
   Srio_osalLog("\n Cleaning Debug Queues\n ");
   count += clean_debug_queues();

    total_descriptors_allocated= NUM_TERMINATE_TX_DESCS +
    		                     NUM_TERMINATE_RX_DESCS +
    		                     NUM_CREDIT_TX_DESCS +
    		                     NUM_CREDIT_RX_DESCS +
    		                     NUM_FWD_PACKETS_MSMC ;


    if(count != (total_descriptors_allocated)) {
       Srio_osalLog("\n\n ***************** De-Allocation of buffers FAILED: Expected %d, Released %d buffers ***************\n",count,total_descriptors_allocated);
       return -1;
    }
  return 0;
}
/******************************************** Configure all queue descriptors **********************************/
/************************************************************************************************/
/************************************************************************************************/
/***********************************************************************************************/

int configure_descriptors(Qmss_QueueHnd queueHnd,Qmss_QueueHnd ret_queueHnd,Qmss_MemRegion qmssRegion,int num_descs,int mtu_size,int isDDR, unsigned int  *PSINFO) {
  /* Set up the Tx Termination & Tx Free Queue */
   int port;
   Qmss_QueueHnd txFreeQueue;
   Qmss_Queue          queueInfo;
   Cppi_DescCfg        descCfg;
   Cppi_Desc * ptrHostDesc;
   Qmss_QueueHnd       srioTempQueue;
   uint8_t             isAllocated,*ptr_dataBuffer;
   int numAllocated,idx,i;

    queueInfo = Qmss_getQueueNumber(queueHnd);
     /* Initialize the Transmit descriptors. We want all the transmit descriptors to go
       * back to the Transmit Free Queue after transmission. */
      memset(&descCfg,0,sizeof(Cppi_DescCfg));
      descCfg.memRegion                 = qmssRegion;
      descCfg.descNum                   = num_descs ; // 32 per port
      descCfg.destQueueNum              = QMSS_PARAM_NOT_SPECIFIED;
      descCfg.queueType                 = Qmss_QueueType_GENERAL_PURPOSE_QUEUE;
      descCfg.initDesc                  = Cppi_InitDesc_INIT_DESCRIPTOR;
      descCfg.descType                  = Cppi_DescType_HOST;
      descCfg.returnQueue               = Qmss_getQueueNumber(ret_queueHnd);
      descCfg.epibPresent               = Cppi_EPIB_NO_EPIB_PRESENT;
      descCfg.returnPushPolicy          = Qmss_Location_HEAD;
      descCfg.cfg.host.returnPolicy     = Cppi_ReturnPolicy_RETURN_BUFFER;
      descCfg.cfg.host.psLocation       = Cppi_PSLoc_PS_IN_DESC;

      /* Initialize the descriptors and place all of them into the general purpose temporary queue */
      srioTempQueue  = Cppi_initDescriptor (&descCfg, &numAllocated);
      if (srioTempQueue < 0)
      {
    	  printf ("Error: Initializing transmit descriptor failed :%d \n", srioTempQueue);
          return -1;
      }

      /* Pop off all descriptors from the temp Queue and move them to the Transmit Free Queue. */
      for (idx = 0; idx < num_descs; idx++)
      {
          /* Get a host descriptor from the free GP queue. */
          ptrHostDesc = (Cppi_Desc* )QMSS_DESC_PTR(Qmss_queuePop(srioTempQueue));
          if (ptrHostDesc == NULL)
          {
            printf ("Error: Transmit Queue Configuration Failed (idx=%d, num_descs=%d)\n", idx, num_descs);
            return -1;
          }

          if(isDDR) {
            /* Allocate memory for the transmit data buffer.(DDR) */
            ptr_dataBuffer = (uint8_t*)Srio_osalDataBufferMalloc(mtu_size);
          } else {
            /* MSMC:  But, for NOW, Everything is done on DDR, this has to be MSMC */
            ptr_dataBuffer = (uint8_t*)Srio_osalDataBufferMalloc(mtu_size);
          }
          if (ptr_dataBuffer == NULL)
          {
        	  printf ("Error: Transmit Data Buffer Allocation Failed. MtuSize = %d\n",mtu_size);
            return -1;
          }

          memset(ptr_dataBuffer,idx,mtu_size);
          /* Set the data and payload length. */
          Cppi_setData (Cppi_DescType_HOST, (Cppi_Desc*)ptrHostDesc, (uint8_t*)ptr_dataBuffer, mtu_size);
          Cppi_setOriginalBufInfo (Cppi_DescType_HOST, (Cppi_Desc*)ptrHostDesc, (uint8_t*)ptr_dataBuffer, mtu_size);
          Cppi_setPacketLen (Cppi_DescType_HOST, (Cppi_Desc*)ptrHostDesc, mtu_size);

          if(PSINFO!=NULL) {
            Cppi_setPSData (Cppi_DescType_HOST,(Cppi_Desc*)ptrHostDesc, (uint8_t *)PSINFO, 8);
          }

          /* OSAL Hook: Writeback the contents of the descriptor. */
          Srio_osalEndDescriptorAccess ((Srio_DrvHandle)NULL, (void *)NULL, 0);

          Srio_osalLog("\n[%d] Cppi_Desc_Addr: %08x ,  DataBuffer: %08x",idx, ptrHostDesc,ptr_dataBuffer);
          /* Debug Message */
          // Srio_osalLog ("DrvConfig TxFree host descriptor 0x%p Buffer: 0x%p\n", ptrHostDesc, ptr_dataBuffer);

          if(PSINFO!=NULL) { /* Credit POOL . Put some patter now*/
        	  uint32_t *ptr =(uint32_t *)ptrHostDesc;
        	  Srio_osalLog("\n Tx Credit Descrpitor Contents [32-48]: { ");
        	  for(i=8;i<12;i++) {
        	     Srio_osalLog(" %08x",ptr[i]);
        	  }
        	  Srio_osalLog(" }");
          }

          /* Add the packet descriptor to the SRIO Transmit Free Queue. */
          Qmss_queuePushDescSize (queueHnd, (void *)ptrHostDesc, SIZE_HOST_DESC);
      }

      /* We are done with the temporary Queue at this stage. */
      Qmss_queueEmpty(srioTempQueue);

      Qmss_queueClose (srioTempQueue);
     /* Set up Tx Credit Free Queue */
      return 0;
 /*  */
}
/******************************************** Configure all queue descriptors **********************************/
/************************************************************************************************/
/************************************************************************************************/
/***********************************************************************************************/

int configure_all_queue_descriptors(SrioQueues_t *srioQueues,int my_id, int dest_id[]) {

   unsigned int PSINFO[2], port;
   int ret;
   int retval=0;
   /* Set up the Tx Termination & Tx Free Queue */
   Srio_osalLog("\n configure_all_queue_descriptors: Configuring the txFree_terminate_queues");

	for(port=0;port<4;port++) {
        Srio_osalLog("\n configure_all_queue_descriptors: Configuring the txFree_terminate_queues for Port %d",port);
		if ((ret=configure_descriptors(srioQueues->txFree_terminate_queues[port],srioQueues->txFree_terminate_queues[port],qmssRegion_terminate,(NUM_TERMINATE_TX_DESCS/4),SRIO_MAX_MTU,TRUE,NULL)) < 0 ) {
			printf("\n ERROR! Configuring the Queue: %d",srioQueues->txFree_terminate_queues[port]);
	      retval=-1;
       }
	}

   /* Set up the Rx terminate Free Queue */
   Srio_osalLog("\n configure_all_queue_descriptors: Configuring the rxFree_terminate_queue");
   if((ret=configure_descriptors(srioQueues->rxFree_terminate_queue,srioQueues->rxFree_terminate_queue,qmssRegion_terminate,(NUM_TERMINATE_RX_DESCS),SRIO_MAX_MTU,TRUE,NULL)) < 0 ) {
	   printf("\n ERROR! Configuring the Queue: %d",srioQueues->rxFree_terminate_queue);
		      retval=-1;
	   }


   /* Set up the Rx Credit Free Queue */
	for(port=0;port<4;port++) {
		Srio_osalLog("\n configure_all_queue_descriptors: Configuring the rxFree_credit_queues for Port %d",port);
		if ((ret=configure_descriptors(srioQueues->rxFree_credit_queues[port],srioQueues->rxFree_credit_queues[port],qmssRegion_credit,(NUM_CREDIT_RX_DESCS/4),SRIO_MAX_CREDIT_MTU,TRUE,NULL)) < 0 ) {
			printf("\n ERROR! Configuring the Queue: %d",srioQueues->rxFree_credit_queues[port]);
	      retval=-1;
       }
	}

	/* Create the SRIO Protocol Specific Information: We always support MAX Packet Size & infinite retries. */

   /* Set up the Tx Credit Queue */
	for(port=0;port<4;port++) {

		CSL_SRIO_SetType11PSInfo(PSINFO, dest_id[port], my_id,/* Seperate Mailbox for credits */
				SRIO_MBOX_CREDIT, SRIO_LETTER_CREDIT, 0, 0xA, 0x0); /* Mailbox 2, letter 3 , 0xA -> 16bytes per segment */
		Srio_osalLog("\n configure_all_queue_descriptors: Configuring the txFree_credit_queues for Port %d",port);
		if ((ret=configure_descriptors(srioQueues->txFree_credit_queues[port],srioQueues->txFree_credit_queues[port],qmssRegion_credit,(NUM_CREDIT_TX_DESCS/4),SRIO_MAX_CREDIT_MTU,TRUE,PSINFO)) < 0 ) {
			printf("\n ERROR! Configuring the Queue: %d",srioQueues->txFree_credit_queues[port]);
	      retval=-1;
       }
	}

   /* Set up Rx Forwarding Queue */
	for(port=0;port<4;port++) {

		CSL_SRIO_SetType11PSInfo(PSINFO, dest_id[port], my_id,/* Seperate Mailbox for forwards */
				SRIO_MBOX_FWD, SRIO_LETTER_FWD, 0, 0xe, 0x0); /* Mailbox 2, letter 3 , 0xe -> 256 bytes per segment */
		Srio_osalLog("\n configure_all_queue_descriptors: Configuring the rxFree_forward_queues for Port %d",port);
		/* For this queue descriptor, make sure it poitns to the rx_forward_return_queue. This is different from the above queues, where
		 * the descriptor pool themselves are the return location as well */
		if ((ret=configure_descriptors(srioQueues->rxFree_forward_queues[port],srioQueues->rx_forward_return_queues[port],qmssRegion_forward,NUM_FWD_POOL_PER_PORT,SRIO_MAX_MTU,TRUE,PSINFO)) < 0 ) {
			printf("\n ERROR! Configuring the Queue: %d",srioQueues->rxFree_forward_queues[port]);
	      retval=-1;
       }
	}
    return (retval);
	/*  */


}
#if 0
void spoofTxCredit(SrioQueues_t *srioQueues) {

	Qmss_QueueHnd           freeQueue;
	Qmss_QueueHnd           freeQueue;

}
#endif
/************************************************************************************************/
/************************************************************************************************/
/***********************************************************************************************/
void PDSP_Config(SrioQueues_t *srioQueues,int my_id,int destIDs[],Srio_PDSPRouterCfg_t *PDSProuterConfig) {

    Qmss_Result             result;
    uint32_t                numAllocated, corenum, startTime, dTime;
    Qmss_QueueHnd           freeQ;
    Qmss_QueueHnd           baseQueue;
    Qmss_QueueHnd           hostRxQueue;
    Qmss_QueueHnd           errorQueue;
    Qmss_QueueHnd           emptyQueue;
    Qmss_QueueHnd           creditPool01;
    Qmss_QueueHnd           creditPool12;
    Qmss_QueueHnd           creditPool3host;
    Qmss_QueueHnd           fwQHnds[QMSS_SRIORTR_FW_QUEUES];
    Qmss_QueueHnd           fakeSrioTxQueue12;
    uint8_t                 isAllocated;
    Qmss_SrioRtrGblCfg      srioRtrGblCfg, srioRtrGblCfgRB;
    Qmss_SrioRtrPortCfg     srioRtrPortCfg[4], srioRtrPortCfgRB;
    Qmss_SrioRtrRouteTbl    routeTbl, routeTblRB;
    Qmss_SrioRtrStats       srioRtrStatsCheck;
    int                     port, i, nextCredit1, nextCredit2;
    unsigned int port_bitmap=0;


    srioRtrGblCfg.queueBase = srioQueues->queueBase;
    srioRtrGblCfg.creditSize = PDSP_CREDIT_PERIOD;
    check_result ("set global config", Qmss_setSrioRtrGblCfg (TEST_PDSP, 0, &srioRtrGblCfg));

    memset (&srioRtrGblCfgRB, 0, sizeof(srioRtrGblCfgRB));
    check_result ("get global config", Qmss_getSrioRtrGblCfg (TEST_PDSP, 0, &srioRtrGblCfgRB));
    check_memory ("check global readback", &srioRtrGblCfg, &srioRtrGblCfgRB, sizeof(srioRtrGblCfg));

    /* Assign */
    for (port = 0; port < 4; port++)
     {
          if(destIDs[port] == -1)
            continue;/* The destination on this port does not exist, so dont configure anything for this port */

    	  srioRtrPortCfg[port].hostSrioTxQueue   = srioQueues->tx_srio_queues[port];

#if A15_INTHE_MIDDLE
   /* ROUTER IN THE MIDDLE */
          if(Role_global==ROLE_ROUTER) {
     	    srioRtrPortCfg[port].creditSrioTxQueue = srioQueues->debugQueues[port];
          } else
#endif
          {
         /* Putting all the credit packets through the same queue */
       	    srioRtrPortCfg[port].creditSrioTxQueue = srioQueues->tx_srio_credit_queues[port];
          }

            for (i = 0; i < 4; i++)
            {
#if 1
            	/* This means that, for every incoming packet on port = 'port', this will be forwared to port 'i', via the queue srioRtrPortCfg[port].sourcePortSrioTxQueue[i] */
            	srioRtrPortCfg[port].sourcePortSrioTxQueue[i] = srioQueues->tx_srio_queues[i];
#else
            	srioRtrPortCfg[port].sourcePortSrioTxQueue[i] = srioQueues->debugQueues[port];
#endif

            }
            /* The destination on this port does not exist, so dont configure anything for this port? */
            if(destIDs[port] == -1) continue;

          srioRtrPortCfg[port].creditReturnQueue = srioQueues->rxFree_credit_queues[port];
          srioRtrPortCfg[port].creditSourceQueue = srioQueues->txFree_credit_queues[port];
          srioRtrPortCfg[port].fwdTxFinCmpQueue = srioQueues->rxFree_forward_queues[port];

          srioRtrPortCfg[port].destID =(my_id << 16) | destIDs[port];/* Destination ID of the remote side, connected to this port */
     }

    /* Configure the ports */
    for (port = 0; port < 4; port++)
    {
        if(destIDs[port] == -1)
       	 continue;/* The destination on this port does not exist, so dont configure anything for this port */

    	check_result ("cfg port", Qmss_setSrioRtrPortCfg (TEST_PDSP, port, 0, &srioRtrPortCfg[port]));
        memset (&srioRtrPortCfgRB, 0, sizeof(srioRtrPortCfgRB));
        check_result ("readback port", Qmss_getSrioRtrPortCfg (TEST_PDSP, port, 0, &srioRtrPortCfgRB));
        check_memory ("check port", &srioRtrPortCfg[port], &srioRtrPortCfgRB, sizeof(srioRtrPortCfgRB));
    }


    /* Configure the PDSP Router */
    memcpy(routeTbl.forwardToPort,PDSProuterConfig->portToForward,256*sizeof(uint8_t));
    memcpy(routeTbl.isFinalDest,PDSProuterConfig->is_dest_field,256*sizeof(uint8_t));

    check_result ("set routing table", Qmss_setSrioRtrGblRouteTbl (TEST_PDSP, 0, &routeTbl));
    check_result ("read routing table", Qmss_getSrioRtrGblRouteTbl (TEST_PDSP, 0, &routeTblRB));
    check_memory ("compare routing table", &routeTbl, &routeTblRB, sizeof(routeTbl));


    /* Enable the PDSP */

    port_bitmap=0;
    for(port=0;port<4;port++) {
      if(destIDs[port]>=0) {
        port_bitmap|=1<<port;
      }
    }
    check_result ("enable ports", Qmss_enableSrioRtrPorts (TEST_PDSP, port_bitmap));
}

/*******************************************************************************************************/
/*******************************************************************************************************/
void printSrioRtrStats ()
{
    Qmss_SrioRtrStats stats;
    int i;

    for (i = 0; i < 4; i++)
    {
        check_result ("printf stats", Qmss_getSrioRtrStats (TEST_PDSP, i, 0, 0, &stats));
	Srio_osalLogC ("stats[%d].hostPkts  = %d\n", i, stats.hostPkts);
	Srio_osalLogC ("stats[%d].packetsRx = %d\n", i, stats.packetsRx);
	Srio_osalLogC ("stats[%d].packetsTx = %d\n", i, stats.packetsTx);
	Srio_osalLogC ("stats[%d].creditsTx = %d\n", i, stats.creditsTx);
	Srio_osalLogC ("stats[%d].creditsRx = %d\n", i, stats.creditsRx);
    }
}
/*******************************************************************************************************/
/*******************************************************************************************************/
static Int32 enable_srio (void)
{
    srioFd = acquireUioGeneric ("srio");
    if (srioFd < 0)
    {
        Srio_osalLog ("Failed to open uio for srio\n");
    }
    return srioFd;
}
/*******************************************************************************************************/
/*******************************************************************************************************/
/*** Setup SRIO communication ***/
int32_t SrioDevice_init_lite_machu (
    CSL_SrioHandle hSrio,
    uint32_t       srioSerdesVAddr,
    int my_id
)
{
    int32_t             i;
    SRIO_PE_FEATURES    peFeatures;
    SRIO_OP_CAR         opCar;
    Qmss_QueueHnd       queueHnd;
    uint8_t             isAllocated;
    uint32_t            gargbageQueue[] = { GARBAGE_LEN_QUEUE,  GARBAGE_TOUT_QUEUE,
                                            GARBAGE_RETRY_QUEUE,GARBAGE_TRANS_ERR_QUEUE,
                                            GARBAGE_PROG_QUEUE, GARBAGE_SSIZE_QUEUE };
    int srioPortNum;
    uint32_t			srioIDMask =0xFF;
    SRIO_TLM_CONTROL    tlmCntlSymbol;
    if (hSrio == NULL)
        return -1;

  Srio_osalLog("\n Starting device init_lite: 1"); fflush(stdout);

    /* TODO: Configure the processing element features
     *  The SRIO RL file is missing the Re-transmit Suppression Support (Bit6) field definition */
    peFeatures.isBridge                          = 0;
    peFeatures.isEndpoint                        = 0;
    peFeatures.isProcessor                       = 1;
    peFeatures.isSwitch                          = 0;
    peFeatures.isMultiport                       = 1;
    peFeatures.isFlowArbiterationSupported       = 0;
    peFeatures.isMulticastSupported              = 0;
    peFeatures.isExtendedRouteConfigSupported    = 0;
    peFeatures.isStandardRouteConfigSupported    = 1;
    peFeatures.isFlowControlSupported            = 1;
    peFeatures.isCRFSupported                    = 0;
    peFeatures.isCTLSSupported                   = 1;
    peFeatures.isExtendedFeaturePtrValid         = 1;
    peFeatures.numAddressBitSupported            = 1;
    CSL_SRIO_SetProcessingElementFeatures (hSrio, &peFeatures);

    Srio_osalLog("\n Starting device init_lite: 2"); fflush(stdout);

    /* Configure the source operation CAR */
    memset ((void *) &opCar, 0, sizeof (opCar));
    opCar.portWriteOperationSupport = 1;
    opCar.atomicClearSupport        = 1;
    opCar.atomicSetSupport          = 1;
    opCar.atomicDecSupport          = 1;
    opCar.atomicIncSupport          = 1;
    opCar.atomicTestSwapSupport     = 1;
    opCar.doorbellSupport           = 1;
    opCar.dataMessageSupport        = 1;
    opCar.writeResponseSupport      = 1;
    opCar.streamWriteSupport        = 1;
    opCar.writeSupport              = 1;
    opCar.readSupport               = 1;
    opCar.dataStreamingSupport      = 1;
    CSL_SRIO_SetSourceOperationCAR (hSrio, &opCar);

    Srio_osalLog("\n Starting device init_lite: 3"); fflush(stdout);

    /* Configure the destination operation CAR */
    memset ((void *) &opCar, 0, sizeof (opCar));
    opCar.portWriteOperationSupport  = 1;
    opCar.doorbellSupport            = 1;
    opCar.dataMessageSupport         = 1;
    opCar.writeResponseSupport       = 1;
    opCar.streamWriteSupport         = 1;
    opCar.writeSupport               = 1;
    opCar.readSupport                = 1;
    CSL_SRIO_SetDestOperationCAR (hSrio, &opCar);

    Srio_osalLog("\n Starting device init_lite: 4"); fflush(stdout);

    /* Enable TLM Base Routing Information for Maintainance Requests & ensure that
        * the BRR's can be used by all the ports. */
    CSL_SRIO_SetDeviceIDCSR (hSrio, my_id, my_id);

      for(srioPortNum=0;srioPortNum<4;srioPortNum++) {
         CSL_SRIO_SetTLMPortBaseRoutingInfo (hSrio, srioPortNum, 1, 1, 1, 0);
         /* Configure the Base Routing Register to ensure that all packets matching the
          * Device Identifier & the Secondary Device Id are admitted. */
         CSL_SRIO_SetTLMPortBaseRoutingPatternMatch (hSrio, srioPortNum, 1, my_id, srioIDMask);
         /* Only add additional route if doing core to core on the same board */
      }
    /* We need to open the Garbage collection queues in the QMSS. This is done to ensure that
     * these queues are not opened by another system entity. */
    for (i = 0; i < 6; i++)
    {
        /* Open the Garabage queues */
        garbageQueueHnd[i] = queueHnd = Qmss_queueOpen (Qmss_QueueType_GENERAL_PURPOSE_QUEUE, gargbageQueue[i], &isAllocated);
        if (queueHnd < 0)
            return -1;

        /* Make sure the queue has not been opened already; we dont the queues to be shared by some other
         * entity in the system. */
        if (isAllocated > 1)
            return -1;
    }
    Srio_osalLog("\n Starting device init_lite: 5"); fflush(stdout);

    /* Set the Transmit Garbage Collection Information. */
    CSL_SRIO_SetTxGarbageCollectionInfo (hSrio, GARBAGE_LEN_QUEUE, GARBAGE_TOUT_QUEUE,
                                         GARBAGE_RETRY_QUEUE, GARBAGE_TRANS_ERR_QUEUE,
                                         GARBAGE_PROG_QUEUE, GARBAGE_SSIZE_QUEUE);
    Srio_osalLog("\n Starting device init_lite: 6"); fflush(stdout);

    /* Configure the PLM for all the ports. */
    for (i = 0; i < 4; i++)
    {
        /* Set the PLM Port Silence Timer. */
        CSL_SRIO_SetPLMPortSilenceTimer (hSrio, i, 0x2);

        /* TODO: We need to ensure that the Port 0 is configured to support both
         * the 2x and 4x modes. The Port Width field is read only. So here we simply
         * ensure that the Input and Output ports are enabled. */
        CSL_SRIO_EnableInputPort (hSrio, i);
        CSL_SRIO_EnableOutputPort (hSrio, i);

        /* Set the PLM Port Discovery Timer. */
        CSL_SRIO_SetPLMPortDiscoveryTimer (hSrio, i, 0x2);

        /* Reset the Port Write Reception capture. */
        CSL_SRIO_SetPortWriteReceptionCapture(hSrio, i, 0x0);
    }

    /* Configuration has been completed. */
    CSL_SRIO_SetBootComplete(hSrio, 1);

    Srio_osalLog("\n Starting device init_lite: 7"); fflush(stdout);

#ifdef MODE_F_ENABLE
  /* For enabling promiscous receive in type 11 */
  {

     unsigned int val;

     for(i=0;i<4;i++) {
       val= hSrio->RIO_TLM[i].RIO_TLM_SP_CONTROL;
       val |= 0x3 << 20;
       hSrio->RIO_TLM[i].RIO_TLM_SP_CONTROL=val;
     }

     val = hSrio->RIO_PER_SET_CNTL;
     val |= 1 << 27;
     hSrio->RIO_PER_SET_CNTL=val;
  }
  Srio_osalLog("\n PERSET_CNTL %x ",hSrio->RIO_PER_SET_CNTL);
  for(i=0;i<4;i++) {
    Srio_osalLog("\n hSrio->RIO_TLM[%d].RIO_TLM_SP_CONTROL= %x",i,hSrio->RIO_TLM[i].RIO_TLM_SP_CONTROL);
  }
#endif
     Srio_osalLog("\n Starting device init_lite: 9"); fflush(stdout);

     /* Initialization has been completed. */
    return 0;
} /* end of Srio_init_lite_machu() */

int32_t SrioDevice_deinit_lite_machu( CSL_SrioHandle hSrio )
{
    int                 i;
    if (hSrio == NULL)
        return -1;
    /* Close the garbage queues */
    for (i = 0; i < 6; i++)
    {
        /* Close the Garabage queues */
        if (Qmss_queueClose (garbageQueueHnd[i]) < QMSS_SOK)
        {
            return -1;
        }

        garbageQueueHnd[i] = QMSS_PARAM_NOT_SPECIFIED;
    }
    return 0;
}
/*************************************************************************************************************/
/*************************************************************************************************************/
int setUP() {
    int status;
    int child = 0;
    int parent = 0;
    int i;
    int retVal = 0;
    off_t pCMemBase;
    Cppi_Handle cppiHnd;
    coreNum = 0;
    parent = 1;
    printf ("coreNum: %d; pid=%d\n", coreNum, getpid());

    fw_osalInit();

      /* Setup RM */
    if (initRm())  {
    	printf ("setupTestFramework: initRm returned error, exiting\n");
        return (-1);
    }
    /* core/task 0 will do cmem alloc and publish to RM; other tasks
     * will query RM */
    pCMemBase = 0;

    if (fw_memAllocInit(&pCMemBase, FW_QMSS_CMEM_SIZE) == fw_FALSE) {
    	printf("ERROR: \"Top Level Test\" fw_memAllocInit failed\n");
        return (-1);
    }

    if (! coreNum)
    {
        fw_name_set ("SRIOMT_CMA_PBASE_LOW", (uint32_t)pCMemBase);
        fw_name_set ("SRIOMT_CMA_PBASE_HIGH", (uint32_t)(pCMemBase >> 32));
    }
#if 0
    {
        uint32_t l, h;
        l = fw_name_lookup ("SRIOMT_CMA_PBASE_LOW");
        h = fw_name_lookup ("SRIOMT_CMA_PBASE_HIGH");
    }
#endif
    /* Create virtual memory maps */
    /* QMSS CFG Regs */
    fw_qmssCfgVaddr = fw_memMap((void*)QMSS_CFG_BASE_ADDR, QMSS_CFG_BLK_SZ);
    if (!fw_qmssCfgVaddr)
    {
    	printf("ERROR: Failed to map QMSS CFG registers\n");
        return (-1);
    }
    Srio_osalLog("main:QMSS_CFG_BASE_ADDR:0x%p Memory mapped at address %p.\n",(void*)QMSS_CFG_BASE_ADDR, fw_qmssCfgVaddr);

    /* QMSS DATA Regs */
    fw_qmssDataVaddr = fw_memMap((void*)QMSS_DATA_BASE_ADDR, QMSS_DATA_BLK_SZ);
    if (!fw_qmssDataVaddr)
    {
    	printf("ERROR: Failed to map QMSS DATA registers\n");
        return (-1);
    }
    Srio_osalLog("main:QMSS_DATA_BASE_ADDR:0x%p Memory mapped at address %p.\n",(void*)QMSS_DATA_BASE_ADDR, fw_qmssDataVaddr);

    /* SRIO CFG Regs */
    fw_srioCfgVaddr = fw_memMap((void*)SRIO_CFG_BASE_ADDR, SRIO_CFG_BLK_SZ);
    if (!fw_srioCfgVaddr)
    {
    	printf("ERROR: Failed to map SRIO registers\n");
        return (-1);
    }
    Srio_osalLog("main:SRIO_CFG_BASE_ADDR:0x%p Memory mapped at address %p.\n",(void*)SRIO_CFG_BASE_ADDR, fw_srioCfgVaddr);

    /* SRIO SERDES CFG Regs */
    fw_srioSerdesCfgVaddr = fw_memMap((void*)SRIO_SERDES_CFG_BASE_ADDR, SRIO_SERDES_CFG_BLK_SZ);
    if (!fw_srioSerdesCfgVaddr)
    {
    	printf("ERROR: Failed to map SRIO SERDES registers\n");
        return (-1);
    }
    Srio_osalLog("main:SRIO_SERDES_CFG_BASE_ADDR:0x%p Memory mapped at address %p.\n",(void*)SRIO_SERDES_CFG_BASE_ADDR, fw_srioSerdesCfgVaddr);

    if(setupTestFramework (&cppiHnd)<0) {
    	printf("\n setupTestFramework Failed !!");
    }

    return retVal;
}
/***********************************************************************************************************************************************/
/******************************************************************************************************************************************************/
/********************************************************************************************************/
/*************************************** multiNodeTestTask ******************************************/
/********************************************************************************************************/

void *srio_drvConfig(int my_id)
{
    Srio_DrvConfig  cfg;
    Srio_DrvHandle  hSrioDriver;
    int port,ret,id;

    /* Initialize the OSAL Data Buffer */
    if (Osal_dataBufferInitMemory(SRIO_MAX_MTU) < 0)
        return;

    /* Initialize the SRIO Driver Configuration. */
    memset ((void *)&cfg, 0, sizeof(Srio_DrvConfig));
    /* Setup the SRIO Driver Managed Configuration. */
    cfg.bAppManagedConfig                   = FALSE;

    /* Driver Managed: Receive Configuration */
    cfg.u.drvManagedCfg.bIsRxCfgValid             = 1;
    cfg.u.drvManagedCfg.rxCfg.rxMemRegion         = qmssRegion_terminate;
    cfg.u.drvManagedCfg.rxCfg.numRxBuffers        = 4;
    cfg.u.drvManagedCfg.rxCfg.rxMTU               = SRIO_MAX_MTU;

    Srio_osalLog("\n Assigning the cfg.u.drvManagedCfg.rxCfg config with RX Queues..");
    /* Pass on Rx Queues for Srio driver */
    cfg.u.drvManagedCfg.rxCfg.rxTerminateQueue  = Sriotype11APIsContext.srioQueues.rx_terminate_queue;
    cfg.u.drvManagedCfg.rxCfg.rxFreeTerminateQueue  = Sriotype11APIsContext.srioQueues.rxFree_terminate_queue;

    for(port=0;port<4;port++) {
      cfg.u.drvManagedCfg.dest_id[port]  = Sriotype11APIsContext.dest_ids_port[port];
      cfg.u.drvManagedCfg.rxCfg.rx_credit_queue[port]  = Sriotype11APIsContext.srioQueues.rx_credit_queues[port];
      cfg.u.drvManagedCfg.rxCfg.rxFree_credit_queue[port]  = Sriotype11APIsContext.srioQueues.rxFree_credit_queues[port];

      cfg.u.drvManagedCfg.rxCfg.rx_forward_queue[port]  = Sriotype11APIsContext.srioQueues.rx_completed_forward_queues[port];
cfg.u.drvManagedCfg.rxCfg.rx_forward_return_queue[port]  = Sriotype11APIsContext.srioQueues.rx_forward_return_queues[port];
      cfg.u.drvManagedCfg.rxCfg.rxFree_forward_queue[port]  = Sriotype11APIsContext.srioQueues.rxFree_forward_queues[port];

    }

    /* Accumulator Configuration is not VALID. */
    cfg.u.drvManagedCfg.rxCfg.bIsAccumlatorCfgValid = 0;

    /* Driver Managed: Transmit Configuration */
    cfg.u.drvManagedCfg.bIsTxCfgValid             = 1;
    cfg.u.drvManagedCfg.txCfg.txMemRegion         = qmssRegion_terminate;
    cfg.u.drvManagedCfg.txCfg.txMTU               = SRIO_MAX_MTU;

    /* Connect to RM */
    cfg.rmServiceHandle = rmClientServiceHandle;

    /* Start the SRIO Driver */
    cfg.srioQueueBase=0;/* Forcing default to SRIO_QUEUE_BASE defined in srio_drv.c*/
    cfg.srioQueueCount=0; /* Forcing default to 16 defined in srio_drv.c */

    cfg.u.drvManagedCfg.my_id = my_id;


    /* The Rx Flow Config amongst the Rx Queues are done inside Srio_start */
    hSrioDriver = Srio_start(&cfg);

    if (hSrioDriver == NULL)
    {
        Srio_osalLog ("Error: SRIO Driver Start Failed\n");
        errorCount++;
        return;
    }

    /* SRIO Driver is operational at this time. */
    Srio_osalLog ("Debug: SRIO Driver has been started Instance Handle %p\n", hSrioDriver);

    /* Hook up the SRIO interrupts with the core. */
    isr_hSrioDriver = hSrioDriver;

    Srio_osalLog("\n Assigning Tx Queues to HW ports through CSL..");
    assignTxSRIOQueuesToHW(fw_srioCfgVaddr,&Sriotype11APIsContext.srioQueues);

    /********************************************* SETUP DONE BY NOW **************************/
    return(isr_hSrioDriver);
}

/**********************************************************************************************/

/*****************************************************************************************/
/**************************** POPULATE PDSP Config ***************************************/
/*****************************************************************************************/
void populate_PDSP_Config(void *base,int my_id,Srio_PDSPRouterCfg_t *PDSProuterConfig) {

	int output_dest_ids[4];
	int i,routing_port;

	memset(PDSProuterConfig->is_dest_field,0,256*sizeof(uint8_t));
    memset(PDSProuterConfig->portToForward,0,256*sizeof(uint8_t));

      Srio_osalLog ("\npopulate_PDSP_config checkpoint 1\n"); fflush(stdout);
	/* Configure PDSP port to forward*/
	for(i=0;i<256;i++) {
	   routing_port=SrioGetRoutingPort(base,my_id,i);
	   if(routing_port!=-1) {
             PDSProuterConfig->portToForward[i]=routing_port;
	     Srio_osalLog("PDSP_RouterConfig: Destination ID %d  forwarded via Port %d\n",i,PDSProuterConfig->portToForward[i]);
	   }
	}

      Srio_osalLog ("\npopulate_PDSP_config checkpoint 2\n"); fflush(stdout);

   	  /* Configure PDSP for only those nodes which are active in the cluster, hence only_active_nodes=TRUE */
      SrioFindMyNeighbours(base,my_id,output_dest_ids,TRUE);
	/* Configure PDSP port to forward*/
	for(i=0;i<4;i++) {
	   Srio_osalLog("SrioFindMyNeighbours: output_dest_ids[%d]=%d\n",i,output_dest_ids[i]);
           if(output_dest_ids[i] !=- 1) {
       	      PDSProuterConfig->is_dest_field[output_dest_ids[i]]=TRUE;
  	      Srio_osalLog("PDSP_RouterConfig.is_dest_field[%d]=TRUE\n",output_dest_ids[i]);
           }
	}
}

/********************************************************************************************************/
/*************************************** DISPLAY ROUTER CONFIG ******************************************/
/********************************************************************************************************/

void printSrioRtrCfg (void)
{
    Qmss_SrioRtrGblCfg      srioRtrGblCfgRB;
    Qmss_SrioRtrRouteTbl    routeTblRB;
    Qmss_SrioRtrPortCfg     srioRtrPortCfgRB;
    int                     i, j;

    /* Print global config */
    memset (&srioRtrGblCfgRB, 0, sizeof(srioRtrGblCfgRB));
    check_result ("printf global config", Qmss_getSrioRtrGblCfg (TEST_PDSP, 0, &srioRtrGblCfgRB));
    Srio_osalLogC ("gblCfg.queueBase  = %d\n", srioRtrGblCfgRB.queueBase);
    Srio_osalLogC ("gblCfg.creditSize = %d\n", srioRtrGblCfgRB.creditSize);

    /* Print each port */
    for (i = 0; i < 4; i++)
    {
        memset (&srioRtrPortCfgRB, 0, sizeof(srioRtrPortCfgRB));
        check_result ("printf port", Qmss_getSrioRtrPortCfg (TEST_PDSP, i, 0, &srioRtrPortCfgRB));
        Srio_osalLogC ("portCfg[%d].hostSrioTxQueue       = %d\n", i, srioRtrPortCfgRB.hostSrioTxQueue);
        Srio_osalLogC ("portCfg[%d].creditSrioTxQueue     = %d\n", i, srioRtrPortCfgRB.creditSrioTxQueue);
        Srio_osalLogC ("portCfg[%d].creditReturnQueue     = %d\n", i, srioRtrPortCfgRB.creditReturnQueue);
        Srio_osalLogC ("portCfg[%d].creditSourceQueue     = %d\n", i, srioRtrPortCfgRB.creditSourceQueue);
        Srio_osalLogC ("portCfg[%d].destID                = 0x%08x\n", i, srioRtrPortCfgRB.destID);
        Srio_osalLogC ("portCfg[%d].fwdTxFinCmpQueue      = %d\n", i, srioRtrPortCfgRB.fwdTxFinCmpQueue);
        Srio_osalLogC ("portCfg[%d].sourcePortSrioTxQueue = %d %d %d %d ", i,
            srioRtrPortCfgRB.sourcePortSrioTxQueue[0], srioRtrPortCfgRB.sourcePortSrioTxQueue[1],
            srioRtrPortCfgRB.sourcePortSrioTxQueue[2], srioRtrPortCfgRB.sourcePortSrioTxQueue[3]);
    }
    /* print routing table */
    check_result ("printf routing table", Qmss_getSrioRtrGblRouteTbl (TEST_PDSP, 0, &routeTblRB));
    for (i = 0; i < 256; i+= 8)
    {
        Srio_osalLog ("Routing[%03d-%03d] = ", i, i+7);
        for (j = 0; j < 8; j++)
        {
            Srio_osalLog ("%d %d, ", routeTblRB.isFinalDest[i+j], routeTblRB.forwardToPort[i+j]);
        }
    }
}
/******************************************************************   APIs ************************************************************************/
/**************************************************************************************************************************************************/
/**************************************************************************************************************************************************/
/* Instantiates SRIO Driver Handle */
/* Invoked at MPI_Init time when it knows its own SRIO ID, but no further information about the neighbors or clusters */
void *Srio_Type11_init (void *baseAddrOfRoutingTbl, int cartridge_id, int node_id)  { /* returns driver handle */

  /* Initializes the SRIO Driver */
  /* Assigns the Dest IDs , outgoing ports */
	    Srio_InitConfig srioInitCfg;
	    Srio_DrvConfig  cfg;
	    int             i,num_remote_entries=0;
	    int port,priority;
	    unsigned short mySRIOId;
	    unsigned short remote_id;
	    int num_cartridge_ids=0;
	    int ret=0;

	    Srio_DrvHandle  srio_drvHandle;
        /* Initialize data structures */

	    Srio_osalLog("\nSrio_Type11_init!! Begin");fflush(stdout);

        for(port=0;port<4;port++) {
           Sriotype11APIsContext.socket_per_port[port]=NULL;
        }
        Sriotype11APIsContext.srioDrvHandle=NULL;



        if(SrioGetSrioIDFromPhysID(baseAddrOfRoutingTbl,cartridge_id,node_id,&mySRIOId)<0) {
           Srio_osalLog("\n ERROR!! Failed to get my own node id ");
      	   return NULL;
        }

        Sriotype11APIsContext.my_id=mySRIOId;
       /* Find my neighbors even if they are not enabled.This is to configure RXU in any case .
        * This step is done at the MPI_init time when no cluster information is available, hence only_active_nodes=FALSE */
	   SrioFindMyNeighbours(baseAddrOfRoutingTbl, Sriotype11APIsContext.my_id, Sriotype11APIsContext.dest_ids_port,FALSE);

	    Srio_osalLog("\nSrio_Type11_init!! Obtained my ID = %d",mySRIOId);fflush(stdout);

	    if(setUP()<0) {
	    	printf("\n FW Setup Failed..\n");
	    	return NULL;
	    }
	    Srio_osalLog("\nSrio_Type11_init!! setUP() done");fflush(stdout);

	    /* Initialize the system only if the core was configured to do so. */
	        Srio_osalLog ("System Initialization for CPPI & QMSS\n");

	        /* System Initialization */
	        if (system_init() < 0) {
		        Srio_osalLog ("system_init() failed\n");
	        	return NULL;
	        }
	        /* Power on SRIO peripheral before using it */
	        if (enable_srio () < 0)
	        {
	            errorCount++;
	            Srio_osalLog ("Error: SRIO PSC Initialization Failed\n");
	            return NULL;
	        }

	        /* Device Specific SRIO Initializations: This should always be called before
	         * initializing the SRIO Driver. */
	        if (SrioDevice_init_lite_machu(fw_srioCfgVaddr, (uint32_t)fw_srioSerdesCfgVaddr, Sriotype11APIsContext.my_id) < 0)
	        {
	            errorCount++;
	            Srio_osalLog ("Error: SrioDevice_init() Initialization Failed\n");
	            return NULL;
	        }

         Srio_osalLog ("Finished srioDeviceInit..\n", coreNum);
	     /* Initialize the SRIO Driver for this task/address space */
	     srioInitCfg.rmServiceHandle = rmClientServiceHandle;
	     srioInitCfg.srioCfgVAddr = fw_srioCfgVaddr;
	     if (Srio_initCfg (&srioInitCfg) < 0)
	     {
	       errorCount++;
	       Srio_osalLog ("Error: SRIO Driver Initialization Failed\n");
	       return NULL;
	     }
	     /* SRIO Driver is operational at this time. */
	     Srio_osalLog ("SRIO Driver has been initialized\n");


	     /* Initialize the OSAL Data Buffer */
	     if (Osal_dataBufferInitMemory(SRIO_MAX_MTU) < 0) {
  	          Srio_osalLog ("Error: Osal_dataBufferInitMemory Failed\n");
	    	 return NULL;
	     }

         /* Create the Queues */
	     Srio_osalLog("\n Allocating Queues..");
	     if((ret=allocate_SRIOQueues(&Sriotype11APIsContext.srioQueues,Sriotype11APIsContext.dest_ids_port)) < 0) {
	     	printf("\n SRIO QUEUE Allocation Failed. Exiting..");
	         return NULL;
	     }
	     Srio_osalLog("\n Configuring Queue descriptors..");
	     /* Configure the queues descriptors */
	     if((ret=configure_all_queue_descriptors(&Sriotype11APIsContext.srioQueues,Sriotype11APIsContext.my_id,Sriotype11APIsContext.dest_ids_port)) < 0) {
	     	printf("\n Configuring Descriptors Failed. Exiting..");
	         return NULL;
	     }

	     display_SRIOQueues(&Sriotype11APIsContext.srioQueues);

	      Srio_osalLog ("\n Finding my_neighbours to configure RXU \n"); fflush(stdout);


	      for(port=0;port<4;port++) {
	          Srio_osalLog("\n Neighbour[Port %d] = %d",i,Sriotype11APIsContext.dest_ids_port[port]);

	     }
	     srio_drvHandle=srio_drvConfig(Sriotype11APIsContext.my_id);

	     if(srio_drvHandle==NULL) {
	    	printf("\n srio_drvConfig Failed!!");
	    	return NULL;
	     }
	     Sriotype11APIsContext.srioDrvHandle=srio_drvHandle;

	     return(srio_drvHandle);
}

/******************************************************************   APIs ************************************************************************/
/**************************************************************************************************************************************************/
/**************************************************************************************************************************************************/
/* Instantiates SRIO Driver Handle */
/* Invoked at MPI Runtime when it knows the cluster information
 * The routing table Algorithm has been already invoked and the baseAddrOfRoutingTbl has them */
int Srio_Type11_setup (void *drvHandle, void *baseAddrOfRoutingTbl, int list_of_srio_ids[], int num_of_srio_ids)
{
  /* Initializes the SRIO Driver */
  /* Assigns the Dest IDs , outgoing ports */
	    int             i,port,val;
        int dest_ids_port;
        Srio_PDSPRouterCfg_t PDSProuterConfig;
        Srio_RuntimeConfig srioDrvRuntimeConfig;
        Srio_SockHandle srioSocket;
        Srio_DrvHandle hSrioDrv = (Srio_DrvHandle )drvHandle;
        CSL_SrioHandle hSrio = (CSL_SrioHandle)fw_srioCfgVaddr;
        Srio_txQueueCfg        txQ;
        Srio_SockBindAddrInfo   bindInfo;

        Srio_osalLog ("\nDJDJBG... About to populate PDSP config\n"); fflush(stdout);

        /* Re-Populate Sriotype11APIsContext.dest_ids_port[] as per the cluster.
         * Please note the cluster has already been defined and baseAddrOfRoutingTbl has been udpated accordingly
         * to filter out which nodes are in the cluster and not */
        SrioFindMyNeighbours(baseAddrOfRoutingTbl,Sriotype11APIsContext.my_id,Sriotype11APIsContext.dest_ids_port,TRUE);

        /* Populate PDSP Config */
        populate_PDSP_Config(baseAddrOfRoutingTbl, Sriotype11APIsContext.my_id, &Sriotype11APIsContext.pdspRouterConfig);

        Srio_osalLog("\n Configuring PDSP ..");
        PDSP_Config(&Sriotype11APIsContext.srioQueues,
        		     Sriotype11APIsContext.my_id,
        		     Sriotype11APIsContext.dest_ids_port,
        		     &Sriotype11APIsContext.pdspRouterConfig);

        Srio_osalLog("\n Configuring PDSP Router Config ..");
        printSrioRtrCfg();

        /****************************  Creating Sockets Per Port ********************************/
        /* Open Sockets for all the adjacent ports */
       for(port=0;port<4;port++)
       {

         if(Sriotype11APIsContext.dest_ids_port[port]<0) /* Port not enabled for this node */
       	 continue;

         /* Open SRIO Socket in Blocking Mode */
         srioSocket =  Srio_sockOpen (hSrioDrv, Srio_SocketType_TYPE11, TRUE);

         if (srioSocket == NULL)
         {
           Srio_osalLog ("Error: Unable to open socket1\n");
           return -1;
         }

         Srio_osalLog ("\n Opening Socket towards Neighbour Id # %d, on port %d\n", Sriotype11APIsContext.dest_ids_port[port],port);

         /* Initialize the core bindings; we use the same mailbox & letter identifiers. */
         memset(&bindInfo,0,sizeof(bindInfo));
         bindInfo.type11.tt       = 0; /* 8 bit ID */
         bindInfo.type11.id       = Sriotype11APIsContext.my_id;
         bindInfo.type11.letter   = SRIO_LETTER_TERMINATE;
         bindInfo.type11.mbox     = SRIO_MBOX_TERMINATE;
         bindInfo.type11.segMap   = 0x1;  /* Multi-segment message */
         bindInfo.type11.remote_id=Sriotype11APIsContext.dest_ids_port[port];
         /* Bind the SRIO socket */
         if (Srio_sockBind (srioSocket , &bindInfo) < 0)
         {
           Srio_osalLog ("Error: socket1 bind failed\n");
             return -1;
         }
         /* Update the Tx Queues */
         memset(&txQ,0,sizeof(Srio_txQueueCfg));

         /* Set the Tx Queue */
         txQ.tx_terminate_queue=Sriotype11APIsContext.srioQueues.tx_terminate_queues[port];
         txQ.txFree_terminate_queue=Sriotype11APIsContext.srioQueues.txFree_terminate_queues[port];

         txQ.tx_srio_queue= Sriotype11APIsContext.srioQueues.tx_srio_queues[port];
         txQ.txSrioChPriority = SRIO_TX_DATA_QUEUE_PRIORITY;

         txQ.tx_credit_queue= Sriotype11APIsContext.srioQueues.tx_srio_credit_queues[port];
         txQ.txCreditChPriority= SRIO_TX_CREDIT_QUEUE_PRIORITY;

         if (Srio_setSockOpt(srioSocket, Srio_Opt_UPDATE_TXQUEUE, &txQ, sizeof(Srio_txQueueCfg)) < 0)
         {
             Srio_osalLog("Error in updating srioSocket txqueue\n");
             return -1;
         }
         Sriotype11APIsContext.socket_per_port[port]= srioSocket;
         srioDrvRuntimeConfig.socket_per_port[port]= srioSocket;
       }

       /* Configuring SRIO Driver's runtime.i.e sockets */
       Srio_runtimeCfg(&srioDrvRuntimeConfig);
       return 0;
}
/******************************************************************   APIs ************************************************************************/
/**************************************************************************************************************************************************/
volatile int djdbg_last_outbound_port = 0;
void *Srio_Type11_AcquireTX (void *drvHandle, void **ptrData, int *buff_len, int srio_dest_id) {
	/* returns cppi_desc */
	Srio_DrvHandle hSrioDrv = (Srio_DrvHandle )drvHandle;
	Srio_SockHandle srioSocket;
	int outbound_port;
    /* Look up the srio_dest_id and find the outbound port number */
	outbound_port= Sriotype11APIsContext.pdspRouterConfig.portToForward[srio_dest_id];
        djdbg_last_outbound_port = outbound_port;
	if(outbound_port==0xff) {
		printf("\n Error! Srio_Type11_AcquireTX: Outbound port not defined for DestID = %d",srio_dest_id);
		return NULL;
	}

    /* Find the socket associaed with the port number */
	srioSocket=Sriotype11APIsContext.socket_per_port[outbound_port];
	if(srioSocket==NULL) {
		printf("\n Error! Srio_Type11_AcquireTX:  Outbound Socket not defined for DestID = %d",srio_dest_id);
		return NULL;
	}
    return(Srio_allocTransmitBuffer(hSrioDrv,(uint8_t**)ptrData,buff_len,srioSocket));

}



/******************************************************************   APIs ************************************************************************/
/**************************************************************************************************************************************************/
int Srio_Type11_SendTX (void *drvHandle, void *cppi_desc, int srio_dest_id, int num_bytes) /* routing tag is already inserted into buffer, first location */
{
	/* returns cppi_desc */
	Srio_DrvHandle hSrioDrv = (Srio_DrvHandle )drvHandle;
	Srio_SockHandle srioSocket;
    Srio_SockAddrInfo       to;
	int outbound_port;
	int my_id,neighbour_dest_id,final_dest_id;


    if ((hSrioDrv == NULL) || (cppi_desc == NULL))
    {
    	printf ("Error ! Srio_Type11_SendTX: (hSrioDrv == NULL) || (cppi_desc == NULL).\n");
       return -1;
    }

	/* Look up the srio_dest_id and find the outbound port number */
	outbound_port= Sriotype11APIsContext.pdspRouterConfig.portToForward[srio_dest_id];

	if(outbound_port==0xff) {
		printf("\n Error! Srio_Type11_SendTX: Outbound port not defined for DestID = %d",srio_dest_id);
		return -1;
	}

   /* Find the socket associaed with the port number */
	srioSocket=Sriotype11APIsContext.socket_per_port[outbound_port];
	if(srioSocket==NULL) {
		printf("\n Error! Srio_Type11_SendTX:  Outbound Socket not defined for DestID = %d",srio_dest_id);
		return -1;
	}


	my_id = Sriotype11APIsContext.my_id;
	neighbour_dest_id = Sriotype11APIsContext.dest_ids_port[outbound_port];

	if(neighbour_dest_id<0) {
		printf("\n Error! Srio_Type11_SendTX:  Neighbour doesntn exist, when routing to DestID= # %d was tried",srio_dest_id);
		return -1;
	}

 	final_dest_id=srio_dest_id;

    to.type11.tt       = 0;/* 8 bit ID */
    to.type11.id       = neighbour_dest_id;
    to.type11.final_dest_id = final_dest_id;

    if(final_dest_id==neighbour_dest_id) {
    	to.type11.mbox     = SRIO_MBOX_TERMINATE;
        to.type11.letter   = SRIO_LETTER_TERMINATE;
    } else {
        to.type11.mbox     = SRIO_MBOX_FWD;
        to.type11.letter   = SRIO_LETTER_FWD;
    }
     /* The payload headers,such as id, are filled out already */
     /* Send the data out from the producer core-or-chip to the consumer core-or-chip. */
     if (Srio_sockSend (srioSocket, cppi_desc,num_bytes, &to) < 0)
     {
    	 printf ("Error! Srio_Type11_SendTX: SRIO Socket send failed\n", coreNum);
        return -1;
     }

     return 0;
}

/******************************************************************************************************************************************/
/******************************************************************************************************************************************/
void *Srio_Type11_AcquireRX(void *drvHandle, void **ptr, int *buff_len) /* returns cppi_desc */
{

      return(Srio_AcquireMessage((Srio_DrvHandle)drvHandle,(uint8_t **) ptr, buff_len));

}
/******************************************************************************************************************************************/
/******************************************************************************************************************************************/
void Srio_Type11_ReleaseRX(void *drvHandle, void *cppi_desc)
{
  Srio_MessageRelease(drvHandle, cppi_desc);
  return;
}
/**************************************************************************************************************************************************/
int Srio_Type11_finalize(void *drvHandle)
{
    /* Stop srio */
	Srio_DrvHandle hSrioDriver = (Srio_DrvHandle )drvHandle;
	Srio_SockHandle srioSocket;
	int retVal=0,port;
    Qmss_Result     qmssResult;

    for(port=0;port<4;port++) {
        /* Close the sockets - common for sender() and receiver() */
        srioSocket=Sriotype11APIsContext.socket_per_port[port];
        if(srioSocket!=NULL) {
          Srio_sockClose (srioSocket);
        }
     }

     if (Srio_stop(hSrioDriver) < 0)   {
    	printf ("Error: SRIO Driver Stop Failed\n");
        errorCount++;
        return -1;
    }

    if(deallocate_SRIOQueues(&Sriotype11APIsContext.srioQueues)<0) {
    	printf ("Error: De-allocation of queues failed\n");
           errorCount++;
           return -1;

    }

    /* close/power down srio */
     if (releaseUioGeneric (srioFd) < 0)
     {
     	printf ("failed to release srio fd\n");
         errorCount++;
     }

    if (Srio_close() < 0)
     {
        errorCount++;
        printf ("core %d: Error: Srio_close() Failed\n", coreNum);
        return -1;
     }


    Srio_osalLog ("System deinit()  Remove Qmss regions, cppi close \n", coreNum);
    /* De-initialize the system only if the core was configured to do so. */
    /* Wait for other cores to de-init first */
    /* Delete the started tags, since we are clearly passed that sync point */
    system_deinit();
    /* Delete the rm server names published */
//    ex_rm_name_del ("srio_mt_region");
    /* Exit qmss */
    Srio_osalLog ("Exit QMSS \n");


    if ((qmssResult = Qmss_exit ()))
    {
        errorCount++;
        printf ("Error: exit error code : %d\n", qmssResult);
    }


    if (errorCount == 0)
    {
        Srio_osalLog ("Debug: MultiNode Example successful.!!\n");
    }

     fw_osalshutdown();
     fw_name_del ("SRIOMT_CMA_PBASE_LOW");
     fw_name_del ("SRIOMT_CMA_PBASE_HIGH");
     fw_memAllocExit ();

     if (errorCount)
     {
         retVal = 1;
     }

     if (retVal)
     {
         Srio_osalLog("****FAIL**** (retVal=%d, errorCount=%d\n", retVal, errorCount);
     }
    return 0;
}

